"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
from inmanta.server import server
from inmanta.server.protocol import Server, ServerSlice
from inmanta.server.agentmanager import AgentManager
from tornado import gen

from typing import List, Generator, Any


class InmantaBootloader(object):
    def __init__(self, agent_no_log: bool = False) -> None:
        self.restserver = Server()
        self.agent_no_log = agent_no_log

    def get_server_slice(self) -> server.Server:
        return server.Server(agent_no_log=self.agent_no_log)

    def get_agent_manager_slice(self) -> AgentManager:
        return AgentManager(self.restserver)

    def get_server_slices(self) -> List[ServerSlice]:
        return [self.get_server_slice(), self.get_agent_manager_slice()]

    @gen.coroutine
    def start(self) -> Generator[Any, None, None]:
        for mypart in self.get_server_slices():
            self.restserver.add_slice(mypart)
        yield self.restserver.start()

    @gen.coroutine
    def stop(self) -> Generator[Any, None, None]:
        yield self.restserver.stop()



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

from configparser import RawConfigParser
import copy
import datetime
import enum
import logging
import uuid
import warnings

from motor import motor_tornado
import pymongo
from tornado import gen

from inmanta import const
import hashlib
from inmanta.const import ResourceState
from _collections import defaultdict
from typing import Dict, List, Set


LOGGER = logging.getLogger(__name__)

DBLIMIT = 100000

# TODO: disconnect
# TODO: difference between None and not set


class Field(object):

    def __init__(self, field_type, required=False, unique=False, **kwargs):

        self._field_type = field_type
        self._required = required

        if "default" in kwargs:
            self._default = True
            self._default_value = kwargs["default"]
        else:
            self._default = False
            self._default_value = None

        self._unique = unique

    def get_field_type(self):
        return self._field_type

    field_type = property(get_field_type)

    def is_required(self):
        return self._required

    required = property(is_required)

    def get_default(self):
        return self._default

    default = property(get_default)

    def get_default_value(self):
        return copy.copy(self._default_value)

    default_value = property(get_default_value)

    def is_unique(self):
        return self._unique

    unique = property(is_unique)


ESCAPE_CHARS = {".": "\uff0E", "\\": "\\\\", "$": "\u0024"}
ESCAPE_CHARS_R = {v: k for k, v in ESCAPE_CHARS.items()}


class DataDocument(object):
    """
        A baseclass for objects that represent data in inmanta. The main purpose of this baseclass is to group dict creation
        logic. These documents are not stored in the database
        (use BaseDocument for this purpose). It provides a to_dict method that the inmanta rpc can serialize. You can store
        DataDocument childeren in BaseDocument fields, they will be serialized to dict. However, on retrieval this is not
        performed.
    """

    def __init__(self, **kwargs):
        self._data = kwargs

    def to_dict(self):
        """
            Return a dict representation of this object.
        """
        return self._data


class DocumentMeta(type):
    def __new__(cls, class_name, bases, dct):
        dct["_fields"] = {}
        for name, field in dct.items():
            if isinstance(field, Field):
                dct["_fields"][name] = field

        for base in bases:
            if hasattr(base, "_fields"):
                dct["_fields"].update(base._fields)

        if len(dct["_fields"]) == 0:
            print(class_name, bases, dict)
        return type.__new__(cls, class_name, bases, dct)


class BaseDocument(object, metaclass=DocumentMeta):
    """
        A base document in the mongodb. Subclasses of this document determine collections names. This type is mainly used to
        bundle query methods and generate validate and query methods for optimized DB access. This is not a full ODM.
    """
    id = Field(field_type=uuid.UUID, required=True)

    _coll = None

    @classmethod
    def collection_name(cls):
        """
            Return the name of the collection
        """
        return cls.__name__

    @classmethod
    def set_connection(cls, motor):
        cls._coll = motor[cls.collection_name()]

    @classmethod
    @gen.coroutine
    def create_indexes(cls):
        # first define all unique indexes
        for name, field in cls._fields.items():
            if field.unique:
                yield cls._coll.create_index([(name, pymongo.ASCENDING)], unique=True)

        if hasattr(cls, "__indexes__"):
            for i in cls.__indexes__:
                keys = i["keys"]
                other = i.copy()
                del other["keys"]
                yield cls._coll.create_index(keys, **other)

    def __init__(self, from_mongo=False, **kwargs):
        self.__fields = self._create_dict(from_mongo, kwargs)

    @classmethod
    def _create_dict(cls, from_mongo, kwargs):
        result = {}
        if from_mongo:
            kwargs = BaseDocument._dict_to_value(kwargs)

            if "id" in kwargs:
                raise AttributeError("A mongo document should not contain a field 'id'")

            kwargs["id"] = kwargs["_id"]
            del kwargs["_id"]
        else:
            if "id" in kwargs:
                raise AttributeError("The id attribute is generated per collection by the document class.")

            kwargs["id"] = cls._new_id()

        fields = cls._fields.copy()
        for name, value in kwargs.items():
            if name not in fields:
                raise AttributeError("%s field is not defined for this document %s" % (name, cls.collection_name()))

            if value is None and fields[name].required:
                raise TypeError("%s field is required" % name)

            if from_mongo and issubclass(fields[name].field_type, enum.Enum):
                value = fields[name].field_type[value]

            if value is not None and not (value.__class__ is fields[name].field_type
                                          or isinstance(value, fields[name].field_type)):
                raise TypeError("Field %s should have the correct type (%s instead of %s)" %
                                (name, fields[name].field_type.__name__, type(value).__name__))

            if value is not None:
                result[name] = value

            elif fields[name].default:
                result[name] = fields[name].default_value

            del fields[name]

        for name in list(fields.keys()):
            if fields[name].default:
                result[name] = fields[name].default_value
                del fields[name]

            elif not fields[name].required:
                del fields[name]

        if len(fields) > 0:
            raise AttributeError("%s fields are required." % ", ".join(fields.keys()))

        return result

    @classmethod
    def mongo_to_dict(cls, **kwargs):
        """
            Validate and transform a dict straight from mongo to a serialization that can be send over rpc
        """
        return cls._create_dict(True, kwargs)

    def _get_field(self, name):
        if hasattr(self.__class__, name):
            field = getattr(self.__class__, name)
            if isinstance(field, Field):
                return field

        return None

    def __getattribute__(self, name):
        if name[0] == "_":
            return object.__getattribute__(self, name)

        field = self._get_field(name)
        if field is not None:
            if name in self.__fields:
                return self.__fields[name]
            else:
                return None

        return object.__getattribute__(self, name)

    def __setattr__(self, name, value):
        if name[0] == "_":
            return object.__setattr__(self, name, value)

        field = self._get_field(name)
        if field is not None:
            # validate
            if value is not None and not isinstance(value, field.field_type):
                raise TypeError("Field %s should be of type %s" % (name, field.field_type))

            self.__fields[name] = value
            return

        raise AttributeError(name)

    @classmethod
    def _value_to_dict(cls, value):
        if isinstance(value, enum.Enum):
            return value.name

        if isinstance(value, DataDocument):
            return cls._encode_keys(value.to_dict())

        if isinstance(value, list):
            return [cls._value_to_dict(x) for x in value]

        if isinstance(value, dict):
            return cls._encode_keys(value)

        return value

    # TODO: make this a generator
    @classmethod
    def _encode_keys(cls, data):
        new_data = {}
        for key, value in data.items():
            new_key = key
            for p, s in ESCAPE_CHARS.items():
                new_key = new_key.replace(p, s)

            new_data[new_key] = cls._value_to_dict(value)

        return new_data

    # TODO: make this a generator
    @classmethod
    def _decode_keys(cls, data):
        new_data = {}
        for key, value in data.items():
            new_key = key
            for p, s in ESCAPE_CHARS_R.items():
                new_key = new_key.replace(p, s)

            new_data[new_key] = cls._dict_to_value(value)

        return new_data

    @classmethod
    def _dict_to_value(cls, value):
        if isinstance(value, list):
            return [cls._dict_to_value(x) for x in value]

        if isinstance(value, dict):
            return cls._decode_keys(value)

        return value

    def to_mongo(self):
        return self._to_dict(True)

    def _to_dict(self, mongo_pk=False):
        """
            Return a dict representing the document
        """
        result = {}
        for name, typing in self._fields.items():
            value = None
            if name in self.__fields:
                value = self.__fields[name]

            if typing.required and value is None:
                raise TypeError("%s should have field '%s'" % (self.__name__, name))

            if value is not None:
                if not isinstance(value, typing.field_type):
                    raise TypeError("Value of field %s does not have the correct type" % name)

                if mongo_pk and name == "id":
                    result["_id"] = value
                else:
                    result[name] = self._value_to_dict(value) if mongo_pk else value

            elif typing.default:
                result[name] = self._value_to_dict(typing.default_value) if mongo_pk else value

        return result

    def to_dict(self):
        return self._to_dict()

    @classmethod
    def _new_id(cls):
        """
            Generate a new ID. Override to use something else than uuid4
        """
        return uuid.uuid4()

    @gen.coroutine
    def insert(self):
        """
            Insert a new document based on the instance passed. Validation is done based on the defined fields.
        """

        yield self._coll.insert_one(self.to_mongo())

    @classmethod
    @gen.coroutine
    def insert_many(cls, documents):
        """
            Insert multiple objects at once
        """
        if len(documents) > 0:
            yield cls._coll.insert_many((d.to_mongo() for d in documents))

    @gen.coroutine
    def update(self, **kwargs):
        """
            Update this document in the database. It will update the fields in this object and send a full update to mongodb.
            Use update_fields to only update specific fields.
        """
        for name, value in kwargs.items():
            setattr(self, name, value)

        yield self._coll.replace_one({"_id": self.id}, self.to_mongo())

    @gen.coroutine
    def update_fields(self, **kwargs):
        """
            Update the given fields of this document in the database. It will update the fields in this object and do a specific
            $set in the mongodb on this document.
        """
        items = {}
        for name, value in kwargs.items():
            setattr(self, name, value)
            items[name] = self._value_to_dict(value)

        yield self._coll.update_one({"_id": self.id}, {"$set": items})

    @classmethod
    @gen.coroutine
    def get_by_id(cls, doc_id: uuid.UUID):
        """
            Get a specific document based on its ID

            :return: An instance of this class with its fields filled from the database.
        """
        result = yield cls._coll.find_one({"_id": doc_id})
        if result is not None:
            return cls(from_mongo=True, **result)

    @classmethod
    @gen.coroutine
    def get_list(cls, **query):
        """
            Get a list of documents matching the filter args
        """
        result = []
        cursor = cls._coll.find(query)

        while (yield cursor.fetch_next):
            obj = cls(from_mongo=True, **cursor.next_object())
            result.append(obj)

        return result

    @classmethod
    @gen.coroutine
    def delete_all(cls, **query):
        """
            Delete all documents that match the given query
        """
        result = yield cls._coll.delete_many(query)
        return result.deleted_count

    @gen.coroutine
    def delete(self):
        """
            Delete this document
        """
        yield self._coll.delete_one({"_id": self.id})

    @gen.coroutine
    def delete_cascade(self):
        yield self.delete()

    @classmethod
    @gen.coroutine
    def query(cls, query):
        cursor = cls._coll.find(query)
        objects = []
        while (yield cursor.fetch_next):
            objects.append(cls(from_mongo=True, **cursor.next_object()))

        return objects


class Project(BaseDocument):
    """
        An inmanta configuration project

        :param name The name of the configuration project.
    """
    name = Field(field_type=str, required=True, unique=True)

    @gen.coroutine
    def delete_cascade(self):
        environments = yield Environment.get_list(project=self.id)
        for env in environments:
            yield env.delete_cascade()
        yield self.delete()


def convert_boolean(value):
    if isinstance(value, bool):
        return value

    if value.lower() not in RawConfigParser.BOOLEAN_STATES:
        raise ValueError('Not a boolean: %s' % value)
    return RawConfigParser.BOOLEAN_STATES[value.lower()]


def convert_int(value):
    if isinstance(value, (int, float)):
        return value

    f_value = float(value)
    i_value = int(value)

    if i_value == f_value:
        return i_value
    return f_value


def convert_agent_map(value):
    if not isinstance(value, dict):
        raise ValueError("Agent map should be a dict")

    for key, v in value.items():
        if not isinstance(key, str):
            raise ValueError("The key of an agent map should be string")

        if not isinstance(v, str):
            raise ValueError("The value of an agent map should be string")

    return value


def convert_agent_trigger_method(value):
    if isinstance(value, const.AgentTriggerMethod):
        return value
    value = str(value)
    valid_values = [x.name for x in const.AgentTriggerMethod]
    if value not in valid_values:
        raise ValueError("%s is not a valid agent trigger method. Valid value: %s" % (value, ','.join(valid_values)))
    return value


AUTO_DEPLOY = "auto_deploy"
PUSH_ON_AUTO_DEPLOY = "push_on_auto_deploy"
AGENT_TRIGGER_METHOD_ON_AUTO_DEPLOY = "agent_trigger_method_on_auto_deploy"
AUTOSTART_SPLAY = "autostart_splay"
AUTOSTART_AGENT_DEPLOY_INTERVAL = "autostart_agent_deploy_interval"
AUTOSTART_AGENT_DEPLOY_SPLAY_TIME = "autostart_agent_deploy_splay_time"
AUTOSTART_AGENT_REPAIR_INTERVAL = "autostart_agent_repair_interval"
AUTOSTART_AGENT_REPAIR_SPLAY_TIME = "autostart_agent_repair_splay_time"
AUTOSTART_ON_START = "autostart_on_start"
AUTOSTART_AGENT_MAP = "autostart_agent_map"
AUTOSTART_AGENT_INTERVAL = "autostart_agent_interval"
AGENT_AUTH = "agent_auth"
SERVER_COMPILE = "server_compile"
RESOURCE_ACTION_LOGS_RETENTION = "resource_action_logs_retention"


class Setting(object):
    """
        A class to define a new environment setting.
    """

    def __init__(self, name, typ, default=None, doc=None, validator=None, recompile=False, update_model=False,
                 agent_restart=False):
        """
            :param name: The name of the setting.
            :param type: The type of the value. This type is mainly used for documentation purpose.
            :param default: An optional default value for this setting. When a default is set and the
                            is requested from the database, it will return the default value and also store
                            the default value in the database.
            :param doc: The documentation/help string for this setting
            :param validator: A validation and casting function for input settings.
            :param recompile: Trigger a recompile of the model when a setting is updated?
            :param update_model: Update the configuration model (git pull on project and repos)
            :param agent_restart: Restart autostarted agents when this settings is updated.
        """
        self.typ = typ
        self.default = default
        self.doc = doc
        self.validator = validator
        self.recompile = recompile
        self.update = update_model
        self.agent_restart = agent_restart

    def to_dict(self):
        return {"type": self.typ, "default": self.default, "doc": self.doc, "recompile": self.recompile, "update": self.update,
                "agent_restart": self.agent_restart}


class Environment(BaseDocument):
    """
        A deployment environment of a project

        :param id A unique, machine generated id
        :param name The name of the deployment environment.
        :param project The project this environment belongs to.
        :param repo_url The repository url that contains the configuration model code for this environment
        :param repo_url The repository branch that contains the configuration model code for this environment
        :param settings Key/value settings for this environment
    """
    name = Field(field_type=str, required=True)
    project = Field(field_type=uuid.UUID, required=True)
    repo_url = Field(field_type=str, default="")
    repo_branch = Field(field_type=str, default="")
    settings = Field(field_type=dict, default={})

    _settings = {
        AUTO_DEPLOY: Setting(name=AUTO_DEPLOY, typ="bool", default=False,
                             doc="When this boolean is set to true, the orchestrator will automatically release a new version "
                                 "that was compiled by the orchestrator itself.", validator=convert_boolean),
        PUSH_ON_AUTO_DEPLOY: Setting(name=PUSH_ON_AUTO_DEPLOY, typ="bool", default=False,
                                     doc="Push a new version when it has been autodeployed.",
                                     validator=convert_boolean),
        AGENT_TRIGGER_METHOD_ON_AUTO_DEPLOY: Setting(name=AGENT_TRIGGER_METHOD_ON_AUTO_DEPLOY, typ="str",
                                                     default=const.AgentTriggerMethod.push_full_deploy.name,
                                                     validator=convert_agent_trigger_method,
                                                     doc="The agent trigger method to use when "
                                                         + PUSH_ON_AUTO_DEPLOY + " is enabled"),
        AUTOSTART_SPLAY: Setting(name=AUTOSTART_SPLAY, typ="int", default=10,
                                 doc="[DEPRECATED] Splay time for autostarted agents.", validator=convert_int),
        AUTOSTART_AGENT_DEPLOY_INTERVAL: Setting(name=AUTOSTART_AGENT_DEPLOY_INTERVAL, typ="int", default=600,
                                                 doc="The deployment interval of the autostarted agents.",
                                                 validator=convert_int, agent_restart=True),
        AUTOSTART_AGENT_DEPLOY_SPLAY_TIME: Setting(name=AUTOSTART_AGENT_DEPLOY_SPLAY_TIME, typ="int", default=600,
                                                   doc="The splay time on the deployment interval of the autostarted agents.",
                                                   validator=convert_int, agent_restart=True),
        AUTOSTART_AGENT_REPAIR_INTERVAL: Setting(name=AUTOSTART_AGENT_REPAIR_INTERVAL, typ="int", default=86400,
                                                 doc="The repair interval of the autostarted agents.",
                                                 validator=convert_int, agent_restart=True),
        AUTOSTART_AGENT_REPAIR_SPLAY_TIME: Setting(name=AUTOSTART_AGENT_REPAIR_SPLAY_TIME, typ="int", default=600,
                                                   doc="The splay time on the repair interval of the autostarted agents.",
                                                   validator=convert_int, agent_restart=True),
        AUTOSTART_ON_START: Setting(name=AUTOSTART_ON_START, default=True, typ="bool", validator=convert_boolean,
                                    doc="Automatically start agents when the server starts instead of only just in time."),
        AUTOSTART_AGENT_MAP: Setting(name=AUTOSTART_AGENT_MAP, default={"internal": "local:"}, typ="dict",
                                     validator=convert_agent_map,
                                     doc="A dict with key the name of agents that should be automatically started. The value "
                                     "is either an empty string or an agent map string.", agent_restart=True),
        AUTOSTART_AGENT_INTERVAL: Setting(name=AUTOSTART_AGENT_INTERVAL, default=600, typ="int", validator=convert_int,
                                          doc="[DEPRECATED] Agent interval for autostarted agents in seconds",
                                          agent_restart=True),
        SERVER_COMPILE: Setting(name=SERVER_COMPILE, default=True, typ="bool",
                                validator=convert_boolean, doc="Allow the server to compile the configuration model."),
        RESOURCE_ACTION_LOGS_RETENTION: Setting(name=RESOURCE_ACTION_LOGS_RETENTION, default=7, typ="int",
                                                validator=convert_int, doc="The number of days to retain resource-action logs"),
    }

    _renamed_settings_map = {AUTOSTART_AGENT_DEPLOY_INTERVAL: AUTOSTART_AGENT_INTERVAL,
                             AUTOSTART_AGENT_DEPLOY_SPLAY_TIME: AUTOSTART_SPLAY}  # name new_option -> name deprecated_option

    __indexes__ = [
        dict(keys=[("name", pymongo.ASCENDING), ("project", pymongo.ASCENDING)], unique=True)
    ]

    @gen.coroutine
    def get(self, key):
        """
            Get a setting in this environment.

            :param key: The name/key of the setting. It should be defined in _settings otherwise a keyerror will be raised.
        """
        if key not in self._settings:
            raise KeyError()

        if key in self._renamed_settings_map:
            name_deprecated_setting = self._renamed_settings_map[key]
            if name_deprecated_setting in self.settings and key not in self.settings:
                warnings.warn("Config option %s is deprecated. Use %s instead." % (name_deprecated_setting, key),
                              category=DeprecationWarning)
                return self.settings[name_deprecated_setting]

        if key in self.settings:
            return self.settings[key]

        if self._settings[key].default is None:
            raise KeyError()

        value = self._settings[key].default
        yield self.set(key, value)
        return value

    @gen.coroutine
    def set(self, key, value):
        """
            Set a new setting in this environment.

            :param key: The name/key of the setting. It should be defined in _settings otherwise a keyerror will be raised.
            :param value: The value of the settings. The value should be of type as defined in _settings
        """
        if key not in self._settings:
            raise KeyError()
        if callable(self._settings[key].validator):
            value = self._settings[key].validator(value)
        yield self._coll.update_one({"_id": self.id}, {"$set": {"settings." + key: self._value_to_dict(value)}})
        self.settings[key] = value

    @gen.coroutine
    def unset(self, key):
        """
            Unset a setting in this environment. If a default value is provided, this value will replace the current value.

            :param key: The name/key of the setting. It should be defined in _settings otherwise a keyerror will be raised.
        """
        if key not in self._settings:
            raise KeyError()

        if self._settings[key].default is None:
            yield self._coll.update_one({"_id": self.id}, {"$unset": {"settings." + key: ""}})
            del self.settings[key]
        else:
            yield self.set(key, self._settings[key].default)

    @gen.coroutine
    def delete_cascade(self, only_content=False):
        yield Agent.delete_all(environment=self.id)

        procs = yield AgentProcess.get_list(environment=self.id)
        for proc in procs:
            yield proc.delete_cascade()

        compile_list = yield Compile.get_list(environment=self.id)
        for cl in compile_list:
            yield cl.delete_cascade()

        models = yield ConfigurationModel.get_list(environment=self.id)
        for model in models:
            yield model.delete_cascade()

        yield Parameter.delete_all(environment=self.id)
        yield Form.delete_all(environment=self.id)
        yield FormRecord.delete_all(environment=self.id)

        if not only_content:
            yield self.delete()


SOURCE = ("fact", "plugin", "user", "form", "report")


class Parameter(BaseDocument):
    """
        A parameter that can be used in the configuration model

        :param name The name of the parameter
        :param value The value of the parameter
        :param environment The environment this parameter belongs to
        :param source The source of the parameter
        :param resource_id An optional resource id
        :param updated When was the parameter updated last

        :todo Add history
    """
    name = Field(field_type=str, required=True)
    value = Field(field_type=str, default="", required=True)
    environment = Field(field_type=uuid.UUID, required=True)
    source = Field(field_type=str, required=True)
    resource_id = Field(field_type=str, default="")
    updated = Field(field_type=datetime.datetime)
    metadata = Field(field_type=dict)

    @classmethod
    @gen.coroutine
    def get_updated_before(cls, updated_before):
        cursor = cls._coll.find({"updated": {"$lt": updated_before}})

        params = []
        while (yield cursor.fetch_next):
            params.append(cls(from_mongo=True, **cursor.next_object()))

        return params


class UnknownParameter(BaseDocument):
    """
        A parameter that the compiler indicated that was unknown. This parameter causes the configuration model to be
        incomplete for a specific environment.

        :param name
        :param resource_id
        :param source
        :param environment
        :param version The version id of the configuration model on which this parameter was reported
    """
    name = Field(field_type=str, required=True)
    environment = Field(field_type=uuid.UUID, required=True)
    source = Field(field_type=str, required=True)
    resource_id = Field(field_type=str, default="")
    version = Field(field_type=int, required=True)
    metadata = Field(field_type=dict)
    resolved = Field(field_type=bool, default=False)

    __indexes__ = [
        dict(keys=[("environment", pymongo.ASCENDING), ("version", pymongo.ASCENDING)])
    ]


class AgentProcess(BaseDocument):
    """
        A process in the infrastructure that has (had) a session as an agent.

        :param hostname The hostname of the device.
        :prama environment To what environment is this process bound
        :param last_seen When did the server receive data from the node for the last time.
    """
    hostname = Field(field_type=str, required=True)
    environment = Field(field_type=uuid.UUID, required=True)
    first_seen = Field(field_type=datetime.datetime, default=None)
    last_seen = Field(field_type=datetime.datetime, default=None)
    expired = Field(field_type=datetime.datetime, default=None)
    sid = Field(field_type=uuid.UUID, required=True)

    @classmethod
    @gen.coroutine
    def get_live(cls, environment=None):
        query = {"$or": [{"expired": {"$exists": False}}, {"expired": None}]}
        if environment is not None:
            query["environment"] = environment

        cursor = cls._coll.find(query)
        nodes = yield cursor.to_list(DBLIMIT)
        return [cls(from_mongo=True, **node) for node in nodes]

    @classmethod
    @gen.coroutine
    def get_live_by_env(cls, env):
        result = yield cls.get_live(env)
        return result

    @classmethod
    @gen.coroutine
    def get_by_env(cls, env):
        nodes = yield cls.get_list(environment=env)
        return nodes

    @classmethod
    @gen.coroutine
    def get_by_sid(cls, sid):
        # TODO: unique index?
        cursor = cls._coll.find({"$or": [{"expired": {"$exists": False}}, {"expired": None}], "sid": sid})
        objects = yield cursor.to_list(DBLIMIT)

        if len(objects) == 0:
            return None
        elif len(objects) > 1:
            LOGGER.exception("Multiple objects with the same unique id found!")
            return cls(from_mongo=True, **objects[0])
        else:
            return cls(from_mongo=True, **objects[0])

    @gen.coroutine
    def delete_cascade(self):
        yield AgentInstance.delete_all(process=self.id)
        yield self.delete()


class AgentInstance(BaseDocument):
    """
        A physical server/node in the infrastructure that reports to the management server.

        :param hostname The hostname of the device.
        :param last_seen When did the server receive data from the node for the last time.
    """
    # TODO: add env to speed up cleanup
    process = Field(field_type=uuid.UUID, required=True)
    name = Field(field_type=str, required=True)
    expired = Field(field_type=datetime.datetime)
    tid = Field(field_type=uuid.UUID, required=True)

    @classmethod
    @gen.coroutine
    def active_for(cls, tid, endpoint):
        objects = yield cls.query({"$or": [{"expired": {"$exists": False}}, {"expired": None}], "tid": tid, "name": endpoint})
        return objects

    @classmethod
    @gen.coroutine
    def active(cls):
        objects = yield cls.query({"$or": [{"expired": {"$exists": False}}, {"expired": None}]})
        return objects


class Agent(BaseDocument):
    """
        An inmanta agent

        :param environment The environment this resource is defined in
        :param name The name of this agent
        :param last_failover Moment at which the primary was last changed
        :param paused is this agent paused (if so, skip it)
        :param primary what is the current active instance (if none, state is down)
    """
    environment = Field(field_type=uuid.UUID, required=True)
    name = Field(field_type=str, required=True)
    last_failover = Field(field_type=datetime.datetime)
    paused = Field(field_type=bool, default=False)
    primary = Field(field_type=uuid.UUID)  # AgentInstance

    __indexes__ = [
        dict(keys=[("environment", pymongo.ASCENDING), ("name", pymongo.ASCENDING)], unique=True)
    ]

    def get_status(self):
        if self.paused:
            return "paused"
        if self.primary is not None:
            return "up"
        return "down"

    def to_dict(self):
        base = BaseDocument.to_dict(self)
        if self.last_failover is None:
            base["last_failover"] = ""

        if self.primary is None:
            base["primary"] = ""

        base["state"] = self.get_status()

        return base

    @classmethod
    @gen.coroutine
    def get(cls, env, endpoint):
        obj = yield cls._coll.find_one({"environment": env, "name": endpoint})

        if obj is not None:
            return cls(from_mongo=True, **obj)


class Report(BaseDocument):
    """
        A report of a substep of compilation

        :param started when the substep started
        :param completed when it ended
        :param command the command that was executed
        :param name The name of this step
        :param errstream what was reported on system err
        :param outstream what was reported on system out
    """
    started = Field(field_type=datetime.datetime, required=True)
    completed = Field(field_type=datetime.datetime, required=True)
    command = Field(field_type=str, required=True)
    name = Field(field_type=str, required=True)
    errstream = Field(field_type=str, default="")
    outstream = Field(field_type=str, default="")
    returncode = Field(field_type=int)

    compile = Field(field_type=uuid.UUID)

    __indexes__ = [
        dict(keys=[("compile", pymongo.ASCENDING)])
    ]


class Compile(BaseDocument):
    """
        A run of the compiler

        :param environment The environment this resource is defined in
        :param started Time the compile started
        :param completed Time to compile was completed
        :param reports Per stage reports
    """
    environment = Field(field_type=uuid.UUID, required=True)
    started = Field(field_type=datetime.datetime)
    completed = Field(field_type=datetime.datetime)

    __indexes__ = [
        dict(keys=[("environment", pymongo.ASCENDING), ("started", pymongo.ASCENDING)])
    ]

    @classmethod
    @gen.coroutine
    def get_reports(cls, queryparts, limit, start, end):
        if limit is not None and end is not None:
            cursor = Compile._coll.find(queryparts).sort("started").limit(int(limit))
            models = []
            while (yield cursor.fetch_next):
                models.append(cls(from_mongo=True, **cursor.next_object()))

            models.reverse()
        else:
            cursor = Compile._coll.find(queryparts).sort("started", pymongo.DESCENDING)
            if limit is not None:
                cursor = cursor.limit(int(limit))
            models = []
            while (yield cursor.fetch_next):
                models.append(cls(from_mongo=True, **cursor.next_object()))

        # load the report stages
        result = []
        for model in models:
            dict_model = model.to_dict()
            result.append(dict_model)

        return result

    @classmethod
    @gen.coroutine
    def get_report(cls, compile_id: uuid.UUID) -> "Compile":
        """
            Get the compile and the associated reports from the database
        """
        result = yield cls.get_by_id(compile_id)
        if result is None:
            return None

        dict_model = result.to_dict()
        cursor = Report._coll.find({"compile": result.id})

        dict_model["reports"] = []
        while (yield cursor.fetch_next):
            obj = Report(from_mongo=True, **cursor.next_object())
            dict_model["reports"].append(obj.to_dict())

        return dict_model

    @gen.coroutine
    def delete_cascade(self):
        yield Report.delete_all(compile=self.id)
        yield self.delete()


class Form(BaseDocument):
    """
        A form in the dashboard defined by the configuration model
    """
    environment = Field(field_type=uuid.UUID, required=True)
    form_type = Field(field_type=str, required=True)
    options = Field(field_type=dict)
    fields = Field(field_type=dict)
    defaults = Field(field_type=dict)
    field_options = Field(field_type=dict)

    @classmethod
    @gen.coroutine
    def get_form(cls, environment, form_type):
        """
            Get a form based on its typed and environment
        """
        forms = yield cls.get_list(environment=environment, form_type=form_type)
        if len(forms) == 0:
            return None
        else:
            return forms[0]


class FormRecord(BaseDocument):
    """
        A form record
    """
    form = Field(field_type=uuid.UUID, required=True)
    environment = Field(field_type=uuid.UUID, required=True)
    fields = Field(field_type=dict)
    changed = Field(field_type=datetime.datetime)


class LogLine(DataDocument):

    @property
    def msg(self):
        return self._data["msg"]

    @property
    def args(self):
        return self._data["args"]

    def get_log_level_as_int(self):
        return self._data["level"].value

    def write_to_logger(self, logger):
        logger.log(self.get_log_level_as_int(), self.msg, *self.args)

    @classmethod
    def log(cls, level, msg, timestamp=None, **kwargs):
        if timestamp is None:
            timestamp = datetime.datetime.now()

        log_line = msg % kwargs
        return cls(level=const.LogLevel(level), msg=log_line, args=[], kwargs=kwargs, timestamp=timestamp)


class ResourceAction(BaseDocument):
    """
        Log related to actions performed on a specific resource version by Inmanta.

        :param resource_version The resource on which the actions are performed
        :param environment The environment this action belongs to.
        :param action_id This is id distinguishes action from each other. Action ids have to be unique per environment.
        :param action The action performed on the resource
        :param started When did the action start
        :param finished When did the action finish
        :param messages The log messages associated with this action
        :param status The status of the resource when this action was finished
        :param changes A dict with key the resource id and value a dict of fields -> value. Value is a dict that can
                       contain old and current keys and the associated values. An empty dict indicates that the field
                       was changed but not data was provided by the agent.
        :param change The change result of an action
    """
    resource_version_ids = Field(field_type=list, required=True)
    environment = Field(field_type=uuid.UUID, required=True)

    action_id = Field(field_type=uuid.UUID, required=True)
    action = Field(field_type=const.ResourceAction, required=True)

    started = Field(field_type=datetime.datetime, required=True)
    finished = Field(field_type=datetime.datetime)

    messages = Field(field_type=list)
    status = Field(field_type=const.ResourceState)
    changes = Field(field_type=dict)
    change = Field(field_type=const.Change)
    send_event = Field(field_type=bool)

    __indexes__ = [
        dict(keys=[("environment", pymongo.ASCENDING), ("action_id", pymongo.ASCENDING)], unique=True),
        dict(keys=[("environment", pymongo.ASCENDING), ("resource_version_ids", pymongo.ASCENDING),
                   ("started", pymongo.DESCENDING)]),
    ]

    def __init__(self, from_mongo=False, **kwargs):
        super().__init__(from_mongo, **kwargs)
        self._updates = {}

    @classmethod
    @gen.coroutine
    def get_log(cls, environment, resource_version_id, action=None, limit=0):
        if action is not None:
            cursor = cls._coll.find({"environment": environment, "resource_version_ids": resource_version_id,
                                     "action": action}).sort("started", direction=pymongo.DESCENDING)
        else:
            cursor = cls._coll.find({"environment": environment, "resource_version_ids": resource_version_id
                                     }).sort("started", direction=pymongo.DESCENDING)

        if limit is not None and limit > 0:
            cursor = cursor.limit(limit)

        log = []
        while (yield cursor.fetch_next):
            log.append(cls(from_mongo=True, **cursor.next_object()))

        return log

    @classmethod
    @gen.coroutine
    def get(cls, environment, action_id):
        resources = yield ResourceAction.get_list(environment=environment, action_id=action_id)
        if len(resources) == 0:
            return None
        return resources[0]

    def set_field(self, name, value):
        if "$set" not in self._updates:
            self._updates["$set"] = {}

        self._updates["$set"][name] = self._value_to_dict(value)

    def add_logs(self, messages):
        if "$push" not in self._updates:
            self._updates["$push"] = {}

        self._updates["$push"]["messages"] = {"$each": self._value_to_dict(messages)}

    def add_changes(self, changes):
        if "$set" not in self._updates:
            self._updates["$set"] = {}

        for resource, values in changes.items():
            for field, change in values.items():
                self._updates["$set"]["changes.%s.%s" % (resource, field)] = self._value_to_dict(change)

    @gen.coroutine
    def save(self):
        """
            Save the accumulated changes
        """
        query = {"environment": self.environment, "action_id": self.action_id}
        if len(self._updates) > 0:
            yield ResourceAction._coll.update_one(query, self._updates)
            self._updates = {}

    @classmethod
    @gen.coroutine
    def purge_logs(cls):
        environments = yield Environment.get_list()
        for env in environments:
            time_to_retain_logs = yield env.get(RESOURCE_ACTION_LOGS_RETENTION)
            keep_logs_until = datetime.datetime.now() - datetime.timedelta(days=time_to_retain_logs)
            yield cls._coll.delete_many({"started": {"$lt": keep_logs_until}})


class Resource(BaseDocument):
    """
        A specific version of a resource. This entity contains the desired state of a resource.

        :param environment: The environment this resource version is defined in
        :param rid: The id of the resource and its version
        :param resource: The resource for which this defines the state
        :param model: The configuration model (versioned) this resource state is associated with
        :param attributes: The state of this version of the resource
        :param attribute_hash: hash of the attributes, excluding requires, provides and version,
                                used to determine if a resource describes the same state across versions
    """
    environment = Field(field_type=uuid.UUID, required=True)
    model = Field(field_type=int, required=True)

    # ID related
    resource_id = Field(field_type=str, required=True)
    resource_version_id = Field(field_type=str, required=True)

    resource_type = Field(field_type=str, required=True)
    agent = Field(field_type=str, required=True)
    id_attribute_name = Field(field_type=str, required=True)
    id_attribute_value = Field(field_type=str, required=True)

    # Field based on content from the resource actions
    last_deploy = Field(field_type=datetime.datetime)

    # State related
    attributes = Field(field_type=dict)
    attribute_hash = Field(field_type=str)
    status = Field(field_type=const.ResourceState, default=const.ResourceState.available)

    # internal field to handle cross agent dependencies
    # if this resource is updated, it must notify all RV's in this list
    # the list contains full rv id's
    provides = Field(field_type=list, default=[])  # List of resource versions

    __indexes__ = [
        dict(keys=[("environment", pymongo.ASCENDING), ("model", pymongo.ASCENDING), ("agent", pymongo.ASCENDING)]),
        dict(keys=[("environment", pymongo.ASCENDING), ("resource_id", pymongo.ASCENDING)]),
        dict(keys=[("environment", pymongo.ASCENDING), ("resource_version_id", pymongo.ASCENDING)], unique=True),
    ]

    def make_hash(self):
        character = "|".join(sorted([str(k) + "||" + str(v)
                                     for k, v in self.attributes.items() if k not in ["requires", "provides", "version"]]))
        m = hashlib.md5()
        m.update(self.resource_id.encode())
        m.update(character.encode())
        self.attribute_hash = m.hexdigest()

    @classmethod
    @gen.coroutine
    def get_resources(cls, environment, resource_version_ids):
        """
            Get all resources listed in resource_version_ids
        """
        cursor = cls._coll.find({"environment": environment, "resource_version_id": {"$in": resource_version_ids}})
        resources = []
        while (yield cursor.fetch_next):
            resources.append(cls(from_mongo=True, **cursor.next_object()))

        return resources

    @classmethod
    @gen.coroutine
    def get_resources_for_attribute_hash(cls, environment, hashes):
        """
            Get all resources listed in resource_version_ids
        """
        cursor = cls._coll.find({"environment": environment, "attribute_hash": {"$in": hashes}})
        resources = []
        while (yield cursor.fetch_next):
            resources.append(cls(from_mongo=True, **cursor.next_object()))

        return resources

    @gen.coroutine
    def delete_cascade(self):
        yield ResourceAction.delete_all(environment=self.environment, resource_version_ids=self.resource_version_id)
        yield self.delete()

    @classmethod
    @gen.coroutine
    def get_undeployable(cls, environment, version):
        """
            Returns a list of resources with an undeployable state
        """
        cursor = cls._coll.find({"environment": environment, "model": version,
                                 "status": {"$in": [x.name for x in const.UNDEPLOYABLE_STATES]}})
        resources = []
        while (yield cursor.fetch_next):
            resources.append(cls(from_mongo=True, **cursor.next_object()))

        return resources

    @classmethod
    @gen.coroutine
    def get_requires(cls, environment, version, resource_version_id):
        """
            Return all resource that have the given resource_verison_id as requires
        """
        cursor = cls._coll.find({"environment": environment, "model": version,
                                 "attributes.requires": resource_version_id})
        resources = []
        while (yield cursor.fetch_next):
            resources.append(cls(from_mongo=True, **cursor.next_object()))

        return resources

    @classmethod
    @gen.coroutine
    def get_resources_report(cls, environment):
        """
            This method generates a report of all resources in the database, with their latest version, if they are deleted
            and when they are last deployed.
                    return {"id": self.resource_id,
                "id_fields": {"type": self.resource_type,
                              "agent": self.agent,
                              "attribute": self.attribute_name,
                              "value": self.attribute_value,
                              },
                "latest_version": self.version_latest,
                "deployed_version": self.version_deployed,
                "last_deploy": self.last_deploy,
                "holds_state": self.holds_state,
                }
        """
        resources = yield cls._coll.find({"environment": environment}, ["resource_id"]).distinct("resource_id")
        result = []
        for res in resources:
            latest = (yield cls._coll.find({"environment": environment,
                                            "resource_id": res}).sort("model", pymongo.DESCENDING).limit(1).to_list(1))[0]

            if latest["status"] == const.ResourceState.available.name:
                cursor = cls._coll.find({"environment": environment, "resource_id": res,
                                         "status": {"$ne": const.ResourceState.available.name}})
                deployed = (yield cursor.sort("model", pymongo.DESCENDING).limit(1).to_list(1))
                deployed = deployed[0] if deployed else {}
            else:
                deployed = latest

            result.append({"resource_id": res,
                           "resource_type": latest["resource_type"],
                           "agent": latest["agent"],
                           "id_attribute_name": latest["id_attribute_name"],
                           "id_attribute_value": latest["id_attribute_value"],
                           "latest_version": latest["model"],
                           "deployed_version": deployed["model"] if "last_deploy" in deployed else None,
                           "last_deploy": deployed["last_deploy"] if "last_deploy" in deployed else None})

        return result

    @classmethod
    @gen.coroutine
    def get_resources_for_version(cls,
                                  environment,
                                  version,
                                  agent=None,
                                  include_attributes=True,
                                  no_obj=False,
                                  include_undefined=True):
        projection = None
        if not include_attributes:
            projection = {"attributes": False}

        filter = {"environment": environment, "model": version}

        if agent is not None:
            filter["agent"] = agent

        if not include_undefined:
            filter["status"] = {"$nin": [const.ResourceState.undefined.name, const.ResourceState.skipped_for_undefined.name]}

        cursor = cls._coll.find(filter, projection)

        resources = []
        while (yield cursor.fetch_next):
            if no_obj:
                resources.append(cls.mongo_to_dict(**cursor.next_object()))
            else:
                resources.append(cls(from_mongo=True, **cursor.next_object()))

        return resources

    @classmethod
    @gen.coroutine
    def get_resources_for_version_raw(cls,
                                      environment,
                                      version,
                                      projection):
        filter = {"environment": environment, "model": version}
        cursor = cls._coll.find(filter, projection)

        resources = []
        while (yield cursor.fetch_next):
            resources.append(cursor.next_object())
        return resources

    @classmethod
    @gen.coroutine
    def get_latest_version(cls, environment, resource_id):
        cursor = cls._coll.find({"environment": environment,
                                 "resource_id": resource_id}).sort("model", pymongo.DESCENDING).limit(1)
        resource = yield cursor.to_list(1)

        if resource is not None and len(resource) > 0:
            return cls(from_mongo=True, **resource[0])

    @classmethod
    @gen.coroutine
    def get(cls, environment, resource_version_id):
        """
            Get a resource with the given resource version id
        """
        value = yield cls._coll.find_one({"environment": environment, "resource_version_id": resource_version_id})
        if value is not None:
            return cls(from_mongo=True, **value)

    @classmethod
    def new(cls, environment, resource_version_id, **kwargs):
        from inmanta.resources import Id
        vid = Id.parse_id(resource_version_id)

        attr = dict(environment=environment, model=vid.version, resource_id=vid.resource_str(),
                    resource_version_id=resource_version_id, resource_type=vid.entity_type, agent=vid.agent_name,
                    id_attribute_name=vid.attribute, id_attribute_value=vid.attribute_value)
        attr.update(kwargs)

        return cls(**attr)

    @classmethod
    @gen.coroutine
    def get_deleted_resources(cls, environment, current_version, current_resources):
        """
            This method returns all resources that have been deleted from the model and are not yet marked as purged. It returns
            the latest version of the resource from a released model.

            :param environment:
            :param current_version:
            :param current_resources: A set of all resource ids in the current version.
        """
        LOGGER.debug("Starting purge_on_delete queries")

        # get all models that have been released
        models = yield ConfigurationModel._coll.find({"environment": environment, "released": True},
                                                     {"version": True, "_id": False}
                                                     ).sort("version", pymongo.DESCENDING).to_list(DBLIMIT)
        versions = set()
        latest_version = None
        for model in models:
            versions.add(model["version"])
            if latest_version is None:
                latest_version = model["version"]

        LOGGER.debug("  All released versions: %s", versions)
        LOGGER.debug("  Latest released version: %s", latest_version)

        # find all resources in previous versions that have "purge_on_delete" set
        resources = yield cls._coll.find({"model": latest_version, "environment": environment,
                                          "$and": [{"attributes.purge_on_delete": {"$exists": True}},
                                                   {"attributes.purge_on_delete": True}]},
                                         ["resource_id"]).distinct("resource_id")
        LOGGER.debug("  Resource with purge_on_delete true: %s", resources)

        # all resources on current model
        LOGGER.debug("  All resource in current version (%s): %s", current_version, current_resources)

        # determined deleted resources
        deleted = set(resources) - current_resources
        LOGGER.debug("  These resources are no longer present in current model: %s", deleted)

        # filter out resources that should not be purged:
        # 1- resources from versions that have not been deployed
        # 2- resources that are already recorded as purged (purged and deployed)
        should_purge = []
        for deleted_resource in deleted:
            # get the full resource history, and determine the purge status of this resource
            cursor = cls._coll.find({"environment": environment, "model": {"$lt": current_version},
                                     "resource_id": deleted_resource}).sort("model", pymongo.DESCENDING)

            while (yield cursor.fetch_next):
                obj = cursor.next_object()

                # if a resource is part of a released version and it is deployed (this last condition is actually enough
                # at the moment), we have found the last status of the resource. If it was not purged in that version,
                # add it to the should purge list.
                if obj["model"] in versions and obj["status"] == const.ResourceState.deployed.name:
                    if not obj["attributes"]["purged"]:
                        should_purge.append(cls(from_mongo=True, **obj))
                    break

        return should_purge

    @classmethod
    def mongo_to_dict(cls, **kwargs):
        dct = super(Resource, cls).mongo_to_dict(**kwargs)
        dct["id"] = dct["resource_version_id"]
        return dct

    def to_dict(self):
        self.make_hash()
        dct = BaseDocument.to_dict(self)
        dct["id"] = dct["resource_version_id"]
        return dct

    def _to_dict(self, mongo_pk=False):
        self.make_hash()
        return BaseDocument._to_dict(self, mongo_pk=mongo_pk)


class ConfigurationModel(BaseDocument):
    """
        A specific version of the configuration model.

        :param version: The version of the configuration model, represented by a unix timestamp.
        :param environment: The environment this configuration model is defined in
        :param date: The date this configuration model was created
        :param released: Is this model released and available for deployment?
        :param deployed: Is this model deployed?
        :param result: The result of the deployment. Success or error.
        :param status: The deployment status of all included resources
        :param version_info: Version metadata
        :param total: The total number of resources
    """
    version = Field(field_type=int, required=True)
    environment = Field(field_type=uuid.UUID, required=True)
    date = Field(field_type=datetime.datetime)

    released = Field(field_type=bool, default=False)
    deployed = Field(field_type=bool, default=False)
    result = Field(field_type=const.VersionState, default=const.VersionState.pending)
    status = Field(field_type=dict, default={})
    version_info = Field(field_type=dict)

    total = Field(field_type=int, default=0)

    # cached state for release
    undeployable = Field(field_type=list, required=False)
    skipped_for_undeployable = Field(field_type=list, required=False)

    __indexes__ = [
        dict(keys=[("environment", pymongo.ASCENDING), ("version", pymongo.ASCENDING)], unique=True)
    ]

    @property
    def done(self):
        return len(self.status)

    @classmethod
    def mongo_to_dict(cls, **kwargs):
        dct = super(DryRun, cls).mongo_to_dict(**kwargs)
        dct["done"] = len(dct["status"])
        return dct

    def to_dict(self):
        dct = BaseDocument.to_dict(self)
        dct["done"] = self.done
        return dct

    @classmethod
    @gen.coroutine
    def get_version(cls, environment, version):
        """
            Get a specific version
        """
        result = yield cls._coll.find_one({"environment": environment, "version": version})
        if result is not None:
            return cls(from_mongo=True, **result)

        return None

    @classmethod
    @gen.coroutine
    def get_latest_version(cls, environment):
        """
            Get the latest released (most recent) version for the given environment
        """
        cursor = cls._coll.find({"environment": environment, "released": True}).sort("version", pymongo.DESCENDING).limit(1)

        versions = yield cursor.to_list(1)

        if len(versions) == 0:
            return None

        return cls(from_mongo=True, **versions[0])

    @classmethod
    @gen.coroutine
    def get_agents(cls, environment, version):
        """
            Returns a list of all agents that have resources defined in this configuration model
        """
        agents = yield Resource._coll.find({"environment": environment, "model": version},
                                           projection={"_id": False, "agent": True}).distinct("agent")

        return agents

    @classmethod
    @gen.coroutine
    def get_versions(cls, environment, start=0, limit=DBLIMIT):
        """
            Get all versions for an environment ordered descending
        """
        cursor = cls._coll.find({"environment": environment}).sort("version", pymongo.DESCENDING).skip(start).limit(limit)

        versions = []
        while (yield cursor.fetch_next):
            versions.append(cls(from_mongo=True, **cursor.next_object()))

        return versions

    @classmethod
    @gen.coroutine
    def set_ready(cls, environment, version, resource_uuid, resource_id, status):
        """
            Mark a resource as deployed in the configuration model status
        """
        entry_uuid = uuid.uuid5(resource_uuid, resource_id)
        resource_key = "status.%s" % entry_uuid
        yield cls._coll.update_one({"environment": environment, "version": version},
                                   {"$set": {resource_key: {"status": cls._value_to_dict(status), "id": resource_id}}})

    @gen.coroutine
    def delete_cascade(self):
        resources = yield Resource.get_list(environment=self.environment, model=self.version)
        for res in resources:
            yield res.delete_cascade()
        yield UnknownParameter.delete_all(environment=self.environment, version=self.version)
        yield Code.delete_all(environment=self.environment, version=self.version)
        yield DryRun.delete_all(environment=self.environment, model=self.version)
        yield self.delete()

    @gen.coroutine
    def get_undeployable(self):
        """
            Returns a list of resource ids (NOT resource version ids) of resources with an undeployable state
        """
        if self.undeployable is None:
            # Fallback if not cached
            resources = yield Resource.get_undeployable(self.environment, self.version)
            self.undeployable = [resource.resource_id for resource in resources]
            yield ConfigurationModel._coll.update_one({"environment": self.environment, "version": self.version},
                                                      {"$set": {"undeployable": self.undeployable}})
        return self.undeployable

    @gen.coroutine
    def get_skipped_for_undeployable(self):
        """
            Returns a list of resource ids (NOT resource version ids)
            of resources which should get a skipped_for_undeployable state
        """
        if self.skipped_for_undeployable is None:
            undeployable = yield Resource.get_undeployable(self.environment, self.version)

            work = list(undeployable)
            skipped = set()

            while len(work) > 0:
                current = work.pop()
                if current.resource_id in skipped:
                    continue
                skipped.add(current.resource_id)
                others = yield Resource.get_requires(self.environment, self.version, current.resource_version_id)
                work.extend(others)

            # get ids
            undeployable = set([resource.resource_id for resource in undeployable])
            self.skipped_for_undeployable = sorted(list(skipped - undeployable))

            yield ConfigurationModel._coll.update_one({"environment": self.environment, "version": self.version},
                                                      {"$set": {"skipped_for_undeployable": self.skipped_for_undeployable}})
        return self.skipped_for_undeployable

    @gen.coroutine
    def mark_done(self):
        """ mark this deploy as done """
        result = const.VersionState.success
        for state in self.status.values():
            if state["status"] != "deployed":
                result = const.VersionState.failed

        yield self.update_fields(deployed=True, result=result)

    @gen.coroutine
    def get_increment(self):
        """
        Find resources incremented by this version compared to deployment state transitions per resource

        :param negative: find resources not in the increment

        available/skipped/unavailable -> next version
        not present -> increment
        error -> increment
        Deployed and same hash -> not increment
        deployed and different hash -> increment
        """
        projection_a = {
            "resource_version_id": True,
            "resource_id": True,
            "status": True,
            "attribute_hash": True,
            "attributes": True
        }
        projection = {
            "resource_version_id": True,
            "resource_id": True,
            "status": True,
            "attribute_hash": True
        }

        # get resources for agent
        resources = yield Resource.get_resources_for_version_raw(
            self.environment,
            self.version,
            projection_a
        )

        # to increment
        increment = []
        not_incrememt = []
        # todo in this verions
        work = list(r for r in resources)

        # get versions
        cursor = self.__class__._coll.find(
            {"environment": self.environment, "released": True}
        ).sort("version", pymongo.DESCENDING)
        versions = []
        while (yield cursor.fetch_next):
            versions.append(cursor.next_object()["version"])

        for version in versions:
            # todo in next verion
            next = []

            vresources = yield Resource.get_resources_for_version_raw(self.environment, version, projection)
            id_to_resource = {r["resource_id"]: r for r in vresources}

            for res in work:
                # not present -> increment
                if res["resource_id"] not in id_to_resource:
                    increment.append(res)
                    continue

                ores = id_to_resource[res["resource_id"]]

                status = ores["status"]
                # available/skipped/unavailable -> next version
                if status in [ResourceState.available.name,
                              ResourceState.skipped.name,
                              ResourceState.unavailable.name]:
                    next.append(res)

                # error -> increment
                elif status in [ResourceState.failed.name,
                                ResourceState.cancelled.name,
                                ResourceState.deploying.name,
                                ResourceState.processing_events.name]:
                    increment.append(res)

                elif status == ResourceState.deployed.name:
                    if res["attribute_hash"] == ores["attribute_hash"]:
                        #  Deployed and same hash -> not increment
                        not_incrememt.append(res)
                    else:
                        # Deployed and different hash -> increment
                        increment.append(res)
                else:
                    LOGGER.warning("Resource in unexpected state: %s, %s", ores["status"], ores["resource_version_id"])
                    increment.append(res)

            work = next
            if not work:
                break
        if work:
            increment.extend(work)

        negative = [res["resource_version_id"] for res in not_incrememt]

        # patch up the graph
        # 1-include stuff for send-events.
        # 2-adapt requires/provides to get closured set

        outset = set((res["resource_version_id"] for res in increment))  # type: Set[str]
        original_provides = defaultdict(lambda: [])  # type: Dict[str,List[str]]
        send_events = []  # type: List[str]

        # build lookup tables
        for res in resources:
            for req in res["attributes"]["requires"]:
                original_provides[req].append(res["resource_version_id"])
            if "send_event" in res["attributes"] and res["attributes"]["send_event"]:
                send_events.append(res["resource_version_id"])

        # recursively include stuff potentially receiving events from nodes in the increment
        work = list(outset)
        done = set()
        while work:
            current = work.pop()
            if current not in send_events:
                # not sending events, so no receivers
                continue

            if current in done:
                continue
            done.add(current)

            provides = original_provides[current]
            work.extend(provides)
            outset.update(provides)

        return set(outset), negative


class Code(BaseDocument):
    """
        A code deployment

        :param environment The environment this code belongs to
        :param version The version of configuration model it belongs to
        :param sources The source code of plugins (phasing out)  form:
            {code_hash:(file_name, provider.__module__, source_code, [req])}
        :param requires Python requires for the source code above
        :param source_refs file hashes refering to files in the file store
            {code_hash:(file_name, provider.__module__, [req])}
    """
    environment = Field(field_type=uuid.UUID, required=True)
    resource = Field(field_type=str, required=True)
    version = Field(field_type=int, required=True)
    sources = Field(field_type=dict)
    source_refs = Field(field_type=dict)

    __indexes__ = [
        dict(keys=[("environment", pymongo.ASCENDING), ("version", pymongo.ASCENDING), ("resource", pymongo.ASCENDING)])
    ]

    @classmethod
    @gen.coroutine
    def get_version(cls, environment, version, resource):
        codes = yield cls.get_list(environment=environment, version=version, resource=resource)
        if len(codes) == 0:
            return None

        return codes[0]

    @classmethod
    @gen.coroutine
    def get_versions(cls, environment, version):
        codes = yield cls.get_list(environment=environment, version=version)
        return codes


class DryRun(BaseDocument):
    """
        A dryrun of a model version

        :param id The id of this dryrun
        :param environment The environment this code belongs to
        :param model The configuration model
        :param date The date the run was requested
        :param resource_total The number of resources that do a dryrun for
        :param resource_todo The number of resources left to do
        :param resources Changes for each of the resources in the version
    """
    environment = Field(field_type=uuid.UUID, required=True)
    model = Field(field_type=int, required=True)
    date = Field(field_type=datetime.datetime)
    total = Field(field_type=int, default=0)
    todo = Field(field_type=int, default=0)
    resources = Field(field_type=dict, default={})

    __indexes__ = [
        dict(keys=[("environment", pymongo.ASCENDING), ("model", pymongo.DESCENDING)])
    ]

    @classmethod
    @gen.coroutine
    def update_resource(cls, dryrun_id, resource_id, dryrun_data):
        """
            Register a resource update with a specific query that sets the dryrun_data and decrements the todo counter, only
            if the resource has not been saved yet.
        """
        entry_uuid = uuid.uuid5(dryrun_id, resource_id)
        resource_key = "resources.%s" % entry_uuid

        query = {"_id": dryrun_id, resource_key: {"$exists": False}}
        update = {"$inc": {"todo": int(-1)}, "$set": {resource_key: cls._value_to_dict(dryrun_data)}}

        yield cls._coll.update_one(query, update)

    @classmethod
    @gen.coroutine
    def create(cls, environment, model, total, todo):
        obj = cls(environment=environment, model=model, date=datetime.datetime.now(), resources={}, total=total, todo=todo)
        obj.insert()
        return obj

    @classmethod
    def mongo_to_dict(cls, **kwargs):
        dct = super(DryRun, cls).mongo_to_dict(**kwargs)
        resources = {r["id"]: r for r in dct["resources"].values()}
        dct["resources"] = resources
        return dct

    def to_dict(self):
        dict_result = BaseDocument.to_dict(self)
        resources = {r["id"]: r for r in dict_result["resources"].values()}
        dict_result["resources"] = resources
        return dict_result


_classes = [Project, Environment, Parameter, UnknownParameter, AgentProcess, AgentInstance, Agent, Report, Compile, Form,
            FormRecord, Resource, ResourceAction, ConfigurationModel, Code, DryRun]


def use_motor(motor):
    for cls in _classes:
        cls.set_connection(motor)


@gen.coroutine
def create_indexes():
    for cls in _classes:
        yield cls.create_indexes()


def connect(host, port, database):
    client = motor_tornado.MotorClient(host, port)
    db = client[database]

    use_motor(db)



"""
    Copyright 2016 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import time
import asyncio
import inspect


async def retry_limited(fun, timeout):
    async def fun_wrapper():
        if inspect.iscoroutinefunction(fun):
            return (await fun())
        else:
            return fun()

    start = time.time()
    while time.time() - start < timeout and not (await fun_wrapper()):
        await asyncio.sleep(0.1)
    if not (await fun_wrapper()):
        raise AssertionError("Bounded wait failed")


UNKWN = object()


def assert_equal_ish(minimal, actual, sortby=[]):
    if isinstance(minimal, dict):
        for k in minimal.keys():
            assert_equal_ish(minimal[k], actual[k], sortby)
    elif isinstance(minimal, list):
        assert len(minimal) == len(actual), "list not equal %s != %s" % (minimal, actual)
        if len(sortby) > 0:
            def keyfunc(val):
                if not isinstance(val, dict):
                    return val
                key = [str(val[x]) for x in sortby if x in val]
                return '_'.join(key)
            actual = sorted(actual, key=keyfunc)
        for (m, a) in zip(minimal, actual):
            assert_equal_ish(m, a, sortby)
    elif minimal is UNKWN:
        return
    else:
        assert minimal == actual


def assert_graph(graph, expected):
    lines = ["%s: %s" % (f.id.get_attribute_value(), t.id.get_attribute_value()) for f in graph.values() for t in f.requires]
    lines = sorted(lines)

    elines = [x.strip() for x in expected.split("\n")]
    elines = sorted(elines)

    assert elines == lines, (lines, elines)



"""
    Copyright 2018 Inmanta
    Copyright 2013 Roman Kalyakin (Mongobox)

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import os
import tempfile
import copy
import subprocess
import time
import sys
import shutil
import socket


MONGOD_BIN = 'mongod'
DEFAULT_ARGS = [
    # don't flood stdout, we're not reading it
    "--quiet",
    # save the port
    # disable unused.
    "--nounixsocket",
    # use a smaller default file size
    "--smallfiles",
    # journaling on by default in 2.0 and makes it to slow
    # for tests, can causes failures in jenkins
    "--nojournal",
    # disable scripting
    "--noscripting",
]
STARTUP_TIME = 0.4
START_CHECK_ATTEMPTS = 200


def find_executable(executable):
    """
        Scan PATH for an executable.
    """
    for path in os.environ.get('PATH', '').split(os.pathsep):
        path = os.path.abspath(path)
        executable_path = os.path.join(path, executable)
        if os.path.isfile(executable_path):
            return executable_path


class MongoProc(object):
    """
        Class to start and manage a mongodb server process
    """
    def __init__(self, port, mongod_bin=None, log_path=None, db_path=None, prealloc=False, auth=False):
        self.mongod_bin = mongod_bin or find_executable(MONGOD_BIN)
        assert self.mongod_bin, 'Could not find "{}" in system PATH. Make sure you have MongoDB installed.'.format(MONGOD_BIN)

        self.port = port
        self.log_path = log_path or os.devnull
        self.prealloc = prealloc
        self.db_path = db_path
        self.auth = auth

        if self.db_path:
            if os.path.exists(self.db_path) and os.path.isfile(self.db_path):
                raise AssertionError('DB path should be a directory, but it is a file.')

        self.process = None

    def start(self):
        """
            Start MongoDB.

            :return: `True` if instance has been started or `False` if it could not start.
        """
        if self.db_path:
            if not os.path.exists(self.db_path):
                os.mkdir(self.db_path)
            self._db_path_is_temporary = False
        else:
            self.db_path = tempfile.mkdtemp()
            self._db_path_is_temporary = True

        args = copy.copy(DEFAULT_ARGS)
        args.insert(0, self.mongod_bin)

        args.extend(['--dbpath', self.db_path])
        args.extend(['--port', str(self.port)])
        args.extend(['--logpath', self.log_path])

        if self.auth:
            args.append("--auth")

        if not self.prealloc:
            args.append("--noprealloc")

        self.process = subprocess.Popen(
            args,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT
        )

        return self._wait_till_started()

    def stop(self):
        if not self.process:
            return

        # Not sure if there should be more checks for
        # other platforms.
        if sys.platform == 'darwin':
            self.process.kill()
        else:
            os.kill(self.process.pid, 9)
        self.process.wait()

        if self._db_path_is_temporary:
            shutil.rmtree(self.db_path)
            self.db_path = None

        self.process = None

    def running(self):
        return self.process is not None

    def _wait_till_started(self):
        attempts = 0
        while self.process.poll() is None and attempts < START_CHECK_ATTEMPTS:
            attempts += 1
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            try:
                try:
                    s.connect(('localhost', int(self.port)))
                    return True
                except (IOError, socket.error):
                    time.sleep(0.25)
            finally:
                s.close()

        self.stop()
        return False

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *args, **kwargs):
        self.stop()



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import hashlib
import os
import subprocess
import shutil
import sys


try:
    import pwd
    import grp  # @UnresolvedImport
except ImportError:
    pwd = None
    grp = None


try:
    from pwd import getpwnam
except ImportError:
    getpwnam = None

try:
    from grp import getgrnam
except ImportError:
    getgrnam = None


class IOBase(object):
    """
        Base class for an IO module. This class is python2 compatible so IOs that work remote can load this module on python2.
    """
    def __init__(self, uri, config):
        """
            Initialize the IO

            :param uri: The uri used to configure this IO
            :param config: The parsed version of the uri
        """
        self.uri = uri
        self.config = config

    def is_remote(self):
        """
            Are operation executed remote

            :return: Returns true if the io operations are remote.
            :rtype: bool
        """
        raise NotImplementedError()

    def close(self):
        """
            Close any resources
        """

    def __del__(self):
        """
            An agent caches IO instances to reuse them for multiple resources. This method is called when an item is removed
            from the cache, for example when a version in the cache is closed.
        """
        self.close()


class BashIO(IOBase):
    """
        This class provides handler IO methods
    """
    def __init__(self, uri, config, run_as=None):
        super(BashIO, self).__init__(uri, config)
        self.run_as = run_as

    def _run_as_args(self, *args):
        """
            Build the arguments to run the command as the `run_as` user
        """
        if self.run_as is None:
            return list(args)

        else:
            arg_str = subprocess.list2cmdline(args)
            sudo_cmd = ["sudo", "-E"]
            ret = sudo_cmd + ["-u", self.run_as, "sh", "-c", arg_str]
            return ret

    def is_remote(self):
        return False

    def hash_file(self, path):
        cwd = os.curdir
        if not os.path.exists(cwd):
            # When this code is executed with nosetests, curdir does not exist anymore
            cwd = "/"

        result = subprocess.Popen(self._run_as_args("sha1sum", path), stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd)
        data = result.communicate()

        if result.returncode > 0 or len(data[1]) > 0:
            raise FileNotFoundError()

        return data[0].decode("utf-8").strip().split(" ")[0]

    def read(self, path):
        """
            Read in the file in path and return its content as string (UTF-8)
        """
        cwd = os.curdir
        if not os.path.exists(cwd):
            # When this code is executed with nosetests, curdir does not exist anymore
            cwd = "/"

        result = subprocess.Popen(self._run_as_args("cat", path), stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd)
        data = result.communicate()

        if result.returncode > 0 or len(data[1]) > 0:
            raise FileNotFoundError()

        return data[0].decode("utf-8")

    def read_binary(self, path):
        """
            Return the content of the file
        """
        result = subprocess.Popen(self._run_as_args("dd", "if=" + path), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        data = result.communicate()

        if result.returncode > 0:
            raise FileNotFoundError()

        return data[0]

    def run(self, command, arguments=[], env=None, cwd=None, timeout=None):
        """
            Execute a command with the given argument and return the result
        """
        current_env = os.environ.copy()
        if env is not None:
            current_env.update(env)

        cmds = [command] + arguments
        result = subprocess.Popen(self._run_as_args(*cmds),
                                  stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=current_env, cwd=cwd)

        if sys.version_info < (3, 0, 0):
            # TODO timeout is not supported
            data = result.communicate()
        else:
            data = result.communicate(timeout=timeout)

        return (data[0].strip().decode("utf-8"), data[1].strip().decode("utf-8"), result.returncode)

    def file_exists(self, path):
        """
            Check if a given file exists
        """
        result = subprocess.Popen(self._run_as_args("stat", "-t", path), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        result.communicate()

        if result.returncode > 0:
            return False

        return True

    def readlink(self, path):
        """
            Return the target of the path
        """
        result = subprocess.Popen(self._run_as_args("readlink", path), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        data = result.communicate()

        if result.returncode > 0:
            raise FileNotFoundError()

        return data[0].decode("utf-8").strip()

    def symlink(self, source, target):
        """
            Symlink source to target
        """
        result = subprocess.Popen(self._run_as_args("ln", "-s", source, target), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        result.communicate()

        if result.returncode > 0:
            return False

        return True

    def is_symlink(self, path):
        """
            Is the given path a symlink
        """
        result = subprocess.Popen(self._run_as_args("stat", path), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        data = result.communicate()

        if result.returncode > 0:
            raise FileNotFoundError()

        if "symbolic link" in data[0].decode("utf-8").strip():
            return True

        return False

    def file_stat(self, path):
        """
            Do a statcall on a file
        """
        result = subprocess.Popen(self._run_as_args("stat", "-c", "%a %U %G", path),
                                  stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        data = result.communicate()

        if result.returncode > 0:
            raise FileNotFoundError()

        parts = data[0].decode("utf-8").strip().split(" ")
        if len(parts) != 3:
            raise IOError()

        status = {}
        status["owner"] = parts[1]
        status["group"] = parts[2]
        status["permissions"] = int(parts[0])

        return status

    def remove(self, path):
        """
            Remove a file
        """
        result = subprocess.Popen(self._run_as_args("rm", "-f", path), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        result.communicate()

        if result.returncode > 0:
            raise FileNotFoundError()

        return True

    def put(self, path, content):
        """
            Put the given content at the given path in UTF-8
        """
        result = subprocess.Popen(self._run_as_args("dd", "of=" + path), stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                  stdin=subprocess.PIPE)
        result.communicate(input=content)

        if result.returncode > 0:
            raise FileNotFoundError()

        return True

    def chown(self, path, user=None, group=None):
        """
            Change the ownership information
        """
        args = None
        if user is not None and group is not None:
            args = ["chown", "%s:%s" % (user, group)]

        elif user is not None:
            args = ["chown", user]

        elif group is not None:
            args = ["chgrp", group]

        if args is not None:
            args.append(path)
            result = subprocess.Popen(self._run_as_args(*args), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            result.communicate()

            if result.returncode > 0:
                raise Exception("Failed to set %s:%s to %s (return code %d)" % (user, group, path, result.returncode))

    def chmod(self, path, permissions):
        """
            Change the permissions
        """
        result = subprocess.Popen(self._run_as_args("chmod", permissions, path),
                                  stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        result.communicate()

        return result.returncode > 0

    def mkdir(self, path):
        """
            Create a directory
        """
        result = subprocess.Popen(self._run_as_args("mkdir", path),
                                  stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        result.communicate()

        return result.returncode > 0

    def rmdir(self, path):
        """
            Remove a directory
        """
        if path == "/":
            raise Exception("Please do not ask to do rm -rf /")

        if "*" in path:
            raise Exception("Do not use wildward in an rm -rf")

        result = subprocess.Popen(self._run_as_args("rm", "-rf", path),
                                  stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        result.communicate()

        return result.returncode > 0

    def __repr__(self):
        if self.run_as is None:
            return "BashIO"

        else:
            return "BashIO_run_as_%s" % self.run_as

    def __str__(self):
        return repr(self)


class LocalIO(IOBase):
    """
        This class provides handler IO methods
    """
    def is_remote(self):
        """
            Are operation executed remote

            :return: Returns true if the io operations are remote.
            :rtype: bool
        """
        return False

    def hash_file(self, path):
        """
            Return the sha1sum of the file at path

            :param str path: The path of the file to hash the content of
            :return: The sha1sum in a hex string
            :rtype: str
        """
        sha1sum = hashlib.sha1()
        with open(path, 'rb') as f:
            sha1sum.update(f.read())

        return sha1sum.hexdigest()

    def read(self, path):
        """
            Read in the file in path and return its content as string

            :param str path: The path of the file to read.
            :return: The string content of the file
            :rtype: string
        """
        with open(path, "rb") as fd:
            return fd.read().decode("utf-8")

    def read_binary(self, path):
        """
            Read in the file in path and return its content as a bytestring

            :param str path: The path of the file to read.
            :return: The byte content of the file
            :rtype: bytes
        """
        with open(path, "rb") as fd:
            return fd.read()

    def run(self, command, arguments=[], env=None, cwd=None, timeout=None):
        """
            Execute a command with the given argument and return the result

            :param str command: The command to execute.
            :param list arguments: The arguments of the command
            :param dict env: A dictionary with environment variables.
            :param str cwd: The working dir to execute the command in.
            :param int timeout: The timeout for this command. This parameter is ignored if the command is executed remotely with
                                a python 2 interpreter.
            :return: A tuple with (stdout, stderr, returncode)
            :rtype: tuple
        """
        current_env = os.environ.copy()
        if env is not None:
            current_env.update(env)
        if sys.version_info < (3, 0, 0):
            current_env = {}
            for k, v in current_env.items():
                current_env[k.encode()] = str(v).encode()

        cmds = [command] + arguments
        result = subprocess.Popen(cmds, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=current_env, cwd=cwd)

        if sys.version_info < (3, 0, 0):
            # TODO timeout is not supported
            data = result.communicate()
        else:
            data = result.communicate(timeout=timeout)

        return (data[0].strip().decode("utf-8"), data[1].strip().decode("utf-8"), result.returncode)

    def file_exists(self, path):
        """
            Check if a given file exists

            :param str path: The path to check if it exists.
            :return: Returns true if the file exists
            :rtype: bool
        """
        return os.path.exists(path)

    def readlink(self, path):
        """
            Return the target of the path

            :param str path: The symlink to get the target for.
            :return: The target of the symlink
            :rtype: str
        """
        return os.readlink(path)

    def symlink(self, source, target):
        """
            Symlink source to target

            :param str source: Create a symlink of this path to target
            :param str target: The path of the symlink to create
        """
        return os.symlink(source, target)

    def is_symlink(self, path):
        """
            Is the given path a symlink

            :param str path: The path of the symlink
            :return: Returns true if the given path points to a symlink
            :rtype: str
        """
        return os.path.islink(path)

    def file_stat(self, path):
        """
            Do a stat call on a file

            :param str path: The file or direct to stat
            :return: A dict with the owner, group and permissions of the given path
            :rtype: dict[str, str]
        """
        stat_result = os.stat(path)
        status = {}
        status["owner"] = pwd.getpwuid(stat_result.st_uid).pw_name
        status["group"] = grp.getgrgid(stat_result.st_gid).gr_name
        status["permissions"] = int(oct(stat_result.st_mode)[-4:])

        return status

    def remove(self, path):
        """
            Remove a file

            :param str path: The path of the file to remove.
        """
        return os.remove(path)

    def put(self, path, content):
        """
            Put the given content at the given path

            :param str path: The location where to write the file
            :param bytes content: The binarystring content to write to the file.
        """
        with open(path, "wb+") as fd:
            fd.write(content)

    def _get_gid(self, name):
        """Returns a gid, given a group name."""
        # Stolen from the python3 shutil lib
        if getgrnam is None or name is None:
            return None
        try:
            result = getgrnam(name)
        except KeyError:
            result = None
        if result is not None:
            return result[2]
        return None

    def _get_uid(self, name):
        """Returns an uid, given a user name."""
        # Stolen from the python3 shutil lib
        if getpwnam is None or name is None:
            return None
        try:
            result = getpwnam(name)
        except KeyError:
            result = None
        if result is not None:
            return result[2]
        return None

    def chown(self, path, user=None, group=None):
        """
            Change the ownership of a file.

            :param str path: The path of the file or directory to change the ownership of.
            :param str user: The user to change to
            :param str group: The group to change to
        """
        # Stolen from the python3 shutil lib
        if user is None and group is None:
            raise ValueError("user and/or group must be set")

        _user = user
        _group = group

        # -1 means don't change it
        if user is None:
            _user = -1
        # user can either be an int (the uid) or a string (the system username)
        elif not isinstance(user, int):
            _user = self._get_uid(user)
            if _user is None:
                raise LookupError("no such user: {!r}".format(user))

        if group is None:
            _group = -1
        elif not isinstance(group, int):
            _group = self._get_gid(group)
            if _group is None:
                raise LookupError("no such group: {!r}".format(group))

        os.chown(path, _user, _group)

    def chmod(self, path, permissions):
        """
            Change the permissions

            :param str path: The path of the file or directory to change the permission of.
            :param str permissions: An octal string with the permission to set.
        """
        os.chmod(path, int(permissions, 8))

    def mkdir(self, path):
        """
            Create a directory

            :param str path: Create this directory. The parent needs to exist.
        """
        os.mkdir(path)

    def rmdir(self, path):
        """
            Remove a directory

            :param str path: The directory to remove
        """
        shutil.rmtree(path)


if __name__ == '__channelexec__':
    global channel

    if os.getuid() == 0:
        local_io = LocalIO(uri="local:", config={})  # type: IOBase
    else:
        local_io = BashIO(uri="local:", config={}, run_as="root")

    for item in channel:  # NOQA
        if hasattr(local_io, item[0]):
            try:
                method = getattr(local_io, item[0])
                result = method(*item[1], **item[2])
                channel.send(result)  # NOQA
            except Exception as e:
                import traceback
                channel.send({"__type__": "RemoteException", "exception_type": str(e.__class__),  # noqa
                              "exception_string": str(e), "traceback": str(traceback.format_exc())})

        else:
            raise AttributeError("Method %s is not supported" % item[0])



"""
    Copyright 2018 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""


from tornado import gen
from tornado import locks
from tornado import process


from inmanta.config import Config
from inmanta import data
from inmanta.server import protocol, SLICE_AGENT_MANAGER, SLICE_SESSION_MANAGER, SLICE_SERVER
from inmanta.asyncutil import retry_limited
from . import config as server_config

import logging
import os
from datetime import datetime
import time
import sys
import uuid
from inmanta.server.protocol import ServerSlice, SessionListener, SessionManager, ReturnClient
from inmanta.server import config as opt
from inmanta.protocol import encode_token, methods
from inmanta.resources import Id
import asyncio

from typing import Optional, Dict, Any, List, Generator, Tuple
from uuid import UUID
from inmanta.server.server import Server

Apireturn = Generator[Any, Any, Tuple[int, Dict[str, Any]]]
NoneGen = Generator[Any, Any, None]


LOGGER = logging.getLogger(__name__)


agent_lock = locks.Lock()


"""
Model in server         On Agent

+---------------+        +----------+
|               |        |          |
|  ENVIRONMENT  |   +---->  PROC    |
|               |   |    |          |
+------+--------+   |    +----+-----+
       |            |         |
       |            |         |
+------v--------+   |    +----v-------------+
|               |   |    |                  |
|  AGENT        |   |    |   AGENT INSTANCE |
|               |   |    |                  |
+------+--------+   |    +------------------+
       |            |
       |            |
+------v--------+   |
|               |   |
|  SESSION      +---+
|               |
+---------------+


get_resources_for_agent

resource_action_update

dryrun_update

set_parameters
"""


@gen.coroutine
def wait_for_proc_bounded(procs: List[process.Subprocess], timeout: float=1.0) -> NoneGen:
    try:
        yield asyncio.wait_for(
            asyncio.gather(
                *[asyncio.shield(proc.wait_for_exit(raise_error=False)) for proc in procs]
            ),
            timeout)
    except asyncio.TimeoutError:
        LOGGER.warning("Agent processes did not close in time (%s)", procs)


class AgentManager(ServerSlice, SessionListener):
    '''
    This class contains all server functionality related to the management of agents
    '''

    def __init__(self, restserver: protocol.Server, closesessionsonstart: bool=True, fact_back_off: int=None) -> None:
        super(AgentManager, self).__init__(SLICE_AGENT_MANAGER)
        self.restserver = restserver

        if fact_back_off is None:
            fact_back_off = opt.server_fact_resource_block.get()

        self._agent_procs: Dict[UUID, process.Subprocess] = {}  # env uuid -> process.SubProcess

        # back-off timer for fact requests
        self._fact_resource_block: int = fact_back_off
        # per resource time of last fact request
        self._fact_resource_block_set: Dict[str, float] = {}

        # session lock
        self.session_lock = locks.Lock()
        # all sessions
        self.sessions: Dict[UUID, protocol.Session] = {}
        # live sessions
        self.tid_endpoint_to_session: Dict[Tuple[UUID, str], protocol.Session] = {}

        self.closesessionsonstart: bool = closesessionsonstart

        self.running: bool = False

    @gen.coroutine
    def prestart(self, server: protocol.Server) -> NoneGen:
        yield ServerSlice.prestart(self, server)

        preserver = server.get_slice(SLICE_SERVER)
        assert isinstance(preserver, Server)
        self._server: Server = preserver
        self._server_storage: Dict[str, str] = self._server._server_storage

        presession = server.get_slice(SLICE_SESSION_MANAGER)
        assert isinstance(presession, SessionManager)
        presession.add_listener(self)

    def new_session(self, session: protocol.Session) -> None:
        self.add_future(self.register_session(session, datetime.now()))

    def expire(self, session: protocol.Session, timeout: float) -> None:
        self.add_future(self.expire_session(session, datetime.now()))

    def seen(self, session: protocol.Session, endpoint_names: List[str]) -> None:
        if set(session.endpoint_names) != set(endpoint_names):
            LOGGER.warning("Agent endpoint set changed, this should not occur, update ignored (was %s is %s)" %
                           (set(session.endpoint_names), set(endpoint_names)))
        # start async, let it run free
        self.add_future(self.flush_agent_presence(session, datetime.now()))

    # From server
    def get_agent_client(self, tid: uuid.UUID, endpoint: str) -> Optional[ReturnClient]:
        if isinstance(tid, str):
            tid = uuid.UUID(tid)
        key = (tid, endpoint)
        if key in self.tid_endpoint_to_session:
            return self.tid_endpoint_to_session[(tid, endpoint)].get_client()
        return None

    @gen.coroutine
    def start(self) -> NoneGen:
        self.running = True
        self.add_future(self.start_agents())
        if self.closesessionsonstart:
            self.add_future(self.clean_db())

    @gen.coroutine
    def stop(self) -> NoneGen:
        yield super().stop()

        self.running = False
        yield self.terminate_agents()

    # Agent Management
    @gen.coroutine
    def ensure_agent_registered(self, env: data.Environment, nodename: str) -> Generator[Any, Any, data.Agent]:
        """
            Make sure that an agent has been created in the database
        """
        with (yield self.session_lock.acquire()):
            agent = yield data.Agent.get(env.id, nodename)
            if agent is not None:
                return agent
            else:
                agent = yield self.create_default_agent(env, nodename)
                return agent

    @gen.coroutine
    def create_default_agent(self, env: data.Environment, nodename: str) -> Generator[Any, Any, data.Agent]:
        saved = data.Agent(environment=env.id, name=nodename, paused=False)
        yield saved.insert()
        yield self.verify_reschedule(env, [nodename])
        return saved

    @gen.coroutine
    def register_session(self, session: protocol.Session, now: float) -> Generator[Any, Any, data.Agent]:
        with (yield self.session_lock.acquire()):
            tid = session.tid
            sid = session.get_id()
            nodename = session.nodename

            self.sessions[sid] = session

            env = yield data.Environment.get_by_id(tid)
            if env is None:
                LOGGER.warning("The environment id %s, for agent %s does not exist!", tid, sid)

            proc = yield data.AgentProcess.get_by_sid(sid)
            if proc is None:
                proc = data.AgentProcess(hostname=nodename, environment=tid, first_seen=now, last_seen=now, sid=sid)
                yield proc.insert()
            else:
                yield proc.update_fields(last_seen=now)

            for nh in session.endpoint_names:
                LOGGER.debug("New session for agent %s on %s", nh, nodename)
                yield data.AgentInstance(tid=tid, process=proc.id, name=nh).insert()
                # yield session.get_client().set_state(agent=nodename, enabled=False)

            if env is not None:
                yield self.verify_reschedule(env, session.endpoint_names)

    @gen.coroutine
    def expire_session(self, session: protocol.Session, now: float) -> NoneGen:
        if not self.running:
            return
        with (yield self.session_lock.acquire()):
            tid = session.tid
            sid = session.get_id()

            LOGGER.debug("expiring session %s", sid)

            del self.sessions[sid]

            env = yield data.Environment.get_by_id(tid)
            if env is None:
                LOGGER.warning("The environment id %s, for agent %s does not exist!", tid, sid)

            aps = yield data.AgentProcess.get_by_sid(sid=sid)
            if aps is None:
                LOGGER.info("expiring session on none existant process sid:%s", sid)
            else:
                yield aps.update_fields(expired=now)

                instances = yield data.AgentInstance.get_list(process=aps.id)
                for ai in instances:
                    yield ai.update_fields(expired=now)

            if env is not None:
                for endpoint in session.endpoint_names:
                    if ((tid, endpoint) in self.tid_endpoint_to_session
                            and self.tid_endpoint_to_session[(tid, endpoint)] == session):
                        del self.tid_endpoint_to_session[(tid, endpoint)]

                yield self.verify_reschedule(env, session.endpoint_names)

    @gen.coroutine
    def get_environment_sessions(self, env_id: uuid.UUID) -> Generator[Any, Any, List[protocol.Session]]:
        """
            Get a list of all sessions for the given environment id
        """
        sessions = yield data.AgentProcess.get_live_by_env(env_id)

        session_list = []
        for session in sessions:
            if session.sid in self.sessions:
                session_list.append(self.sessions[session.sid])

        return session_list

    @gen.coroutine
    def flush_agent_presence(self, session: protocol.Session, now: float) -> NoneGen:
        tid = session.tid
        sid = session.get_id()

        env = yield data.Environment.get_by_id(tid)
        if env is None:
            LOGGER.warning("The environment id %s, for agent %s does not exist!", tid, sid)
            return

        aps = yield data.AgentProcess.get_by_sid(sid=sid)
        if aps is None:
            LOGGER.warning("No process registered for SID %s", sid)
            return

        yield aps.update_fields(last_seen=now)

    @gen.coroutine
    def verify_reschedule(self, env: data.Environment, enpoints: str) -> NoneGen:
        """
             only call under session lock
        """
        if not self.running:
            return
        tid = env.id
        no_primary = [endpoint for endpoint in enpoints if (tid, endpoint) not in self.tid_endpoint_to_session]
        agents = yield [data.Agent.get(env.id, endpoint) for endpoint in no_primary]
        needswork = [agent for agent in agents if agent is not None and not agent.paused]
        for agent in needswork:
            yield self.reschedule(env, agent)

    @gen.coroutine
    def reschedule(self, env: data.Environment, agent: data.Agent) -> NoneGen:
        """
             only call under session lock
        """
        tid = env.id
        instances = yield data.AgentInstance.active_for(tid, agent.name)

        for instance in instances:
            agent_proc = yield data.AgentProcess.get_by_id(instance.process)
            sid = agent_proc.sid

            if sid not in self.sessions:
                LOGGER.warn("session marked as live in DB, but not found. sid: %s" % sid)
            else:
                yield self._set_primary(env, agent, instance, self.sessions[sid])
                return

        yield agent.update_fields(primary=None, last_failover=datetime.now())

    @gen.coroutine
    def _set_primary(self,
                     env: data.Environment,
                     agent: data.Agent,
                     instance: data.AgentInstance,
                     session: protocol.Session) -> NoneGen:
        LOGGER.debug("set session %s as primary for agent %s in env %s" % (session.get_id(), agent.name, env.id))
        self.tid_endpoint_to_session[(env.id, agent.name)] = session
        yield agent.update_fields(last_failover=datetime.now(), primary=instance.id)
        self.add_future(session.get_client().set_state(agent.name, True))

    @gen.coroutine
    def clean_db(self) -> NoneGen:
        with (yield self.session_lock.acquire()):
            LOGGER.debug("Cleaning server session DB")

            # TODO: do as one query
            procs = yield data.AgentProcess.get_live()
            for proc in procs:
                yield proc.update_fields(expired=datetime.now())

            ais = yield data.AgentInstance.active()
            for ai in ais:
                yield ai.update_fields(expired=datetime.now())

            agents = yield data.Agent.get_list()
            for agent in agents:
                yield agent.update_fields(primary=None)

    # utils
    def _fork_inmanta(self,
                      args: List[str],
                      outfile: Optional[str],
                      errfile: Optional[str],
                      cwd: Optional[str]=None) -> process.Subprocess:
        """
            Fork an inmanta process from the same code base as the current code
        """
        inmanta_path = [sys.executable, "-m", "inmanta.app"]
        # handles can be closed, owned by child process,...
        try:
            outhandle = None
            errhandle = None
            if outfile is not None:
                outhandle = open(outfile, "wb+")
            if errfile is not None:
                errhandle = open(errfile, "wb+")

            # TODO: perhaps show in dashboard?
            return process.Subprocess(inmanta_path + args, cwd=cwd, env=os.environ.copy(), stdout=outhandle, stderr=errhandle)
        finally:
            if outhandle is not None:
                outhandle.close()
            if errhandle is not None:
                errhandle.close()

    # External APIS
    @protocol.handle(methods.get_agent_process, agent_id="id")
    @gen.coroutine
    def get_agent_process(self, agent_id: str) -> Apireturn:
        return (yield self.get_agent_process_report(agent_id))

    @protocol.handle(methods.trigger_agent, agent_id="id", env="tid")
    @gen.coroutine
    def trigger_agent(self, env: UUID, agent_id: str) -> Apireturn:
        raise NotImplementedError()

    @protocol.handle(methods.list_agent_processes)
    @gen.coroutine
    def list_agent_processes(self, environment: Optional[UUID], expired: bool) -> Apireturn:
        if environment is not None:
            env = yield data.Environment.get_by_id(environment)
            if env is None:
                return 404, {"message": "The given environment id does not exist!"}

        tid = environment
        if tid is not None:
            if expired:
                aps = yield data.AgentProcess.get_by_env(tid)
            else:
                aps = yield data.AgentProcess.get_live_by_env(tid)
        else:
            if expired:
                aps = yield data.AgentProcess.get_list()
            else:
                aps = yield data.AgentProcess.get_live()

        processes = []
        for p in aps:
            agent_dict = p.to_dict()
            ais = yield data.AgentInstance.get_list(process=p.id)
            oais = []
            for ai in ais:
                a = ai.to_dict()
                oais.append(a)
            agent_dict["endpoints"] = oais
            processes.append(agent_dict)

        return 200, {"processes": processes}

    @protocol.handle(methods.list_agents, env="tid")
    @gen.coroutine
    def list_agents(self, env: Optional[data.Environment]) -> Apireturn:
        if env is not None:
            tid = env.id
            ags = yield data.Agent.get_list(environment=tid)
        else:
            ags = yield data.Agent.get_list()

        return 200, {"agents": [a.to_dict() for a in ags], "servertime": datetime.now().isoformat(timespec='microseconds')}

    @protocol.handle(methods.get_state, env="tid")
    @gen.coroutine
    def get_state(self, env: data.Environment, sid: uuid.UUID, agent: str) -> Apireturn:
        tid: UUID = env.id
        if isinstance(tid, str):
            tid = uuid.UUID(tid)
        key = (tid, agent)
        if key in self.tid_endpoint_to_session:
            session = self.tid_endpoint_to_session[key]
            if session.id == sid:
                return 200, {"enabled": True}
        return 200, {"enabled": False}

    @gen.coroutine
    def get_agent_process_report(self, apid: uuid.UUID) -> Apireturn:
        ap = yield data.AgentProcess.get_by_id(apid)
        if ap is None:
            return 404, {"message": "The given AgentProcess id does not exist!"}
        sid = ap.sid
        if sid not in self.sessions:
            return 404, {"message": "The given AgentProcess is not live!"}
        client = self.sessions[sid].get_client()
        result = yield client.get_status()
        return result.code, result.get_result()

    # Start/stop agents
    @gen.coroutine
    def _ensure_agents(self,
                       env: data.Environment,
                       agents: List[str],
                       restart: bool=False) -> Generator[Any, Any, bool]:
        """
            Ensure that all agents defined in the current environment (model) and that should be autostarted, are started.

            :param env: The environment to start the agents for
            :param agents: A list of agent names that possibly should be started in this environment.
            :param restart: Restart all agents even if the list of agents is up to date.
        """
        agent_map: Dict[str, str]
        agent_map = yield env.get(data.AUTOSTART_AGENT_MAP)
        agents = [agent for agent in agents if agent in agent_map]
        needsstart = restart
        if len(agents) == 0:
            return False

        with (yield agent_lock.acquire()):
            LOGGER.info("%s matches agents managed by server, ensuring it is started.", agents)
            for agent in agents:
                with (yield self.session_lock.acquire()):
                    myagent = self.get_agent_client(env.id, agent)
                    if myagent is None:
                        needsstart = True

            if needsstart:
                res = yield self.__do_start_agent(agents, env)
                return res
        return False

    @gen.coroutine
    def __do_start_agent(self, agents: List[str], env: data.Environment) -> Generator[Any, Any, bool]:
        """
            Start an agent process for the given agents in the given environment
        """
        agent_map: Dict[str, str]
        agent_map = yield env.get(data.AUTOSTART_AGENT_MAP)
        config: str
        config = yield self._make_agent_config(env, agents, agent_map)

        config_dir = os.path.join(self._server_storage["agents"], str(env.id))
        if not os.path.exists(config_dir):
            os.mkdir(config_dir)

        config_path = os.path.join(config_dir, "agent.cfg")
        with open(config_path, "w+") as fd:
            fd.write(config)

        if not self._server._agent_no_log:
            out: Optional[str] = os.path.join(self._server_storage["logs"], "agent-%s.out" % env.id)
            err: Optional[str] = os.path.join(self._server_storage["logs"], "agent-%s.err" % env.id)
        else:
            out = None
            err = None

        agent_log = os.path.join(self._server_storage["logs"], "agent-%s.log" % env.id)
        proc = self._fork_inmanta(["-vvvv", "--timed-logs", "--config", config_path, "--log-file", agent_log, "agent"],
                                  out, err)

        if env.id in self._agent_procs and self._agent_procs[env.id] is not None:
            LOGGER.debug("Terminating old agent with PID %s", self._agent_procs[env.id].proc.pid)
            self._agent_procs[env.id].proc.terminate()
            yield wait_for_proc_bounded([self._agent_procs[env.id]])

        self._agent_procs[env.id] = proc

        # Wait for an agent to start
        yield retry_limited(lambda: self.get_agent_client(env.id, agents[0]) is not None, 5)
        # yield sleep(2)

        LOGGER.debug("Started new agent with PID %s", proc.proc.pid)
        return True

    @gen.coroutine
    def terminate_agents(self) -> NoneGen:
        for proc in self._agent_procs.values():
            proc.proc.terminate()
        yield wait_for_proc_bounded(self._agent_procs.values())

    @gen.coroutine
    def _make_agent_config(self,
                           env: data.Environment,
                           agent_names: List[str],
                           agent_map: Dict[str, str]) -> Generator[Any, Any, str]:
        """
            Generate the config file for the process that hosts the autostarted agents

            :param env: The environment for which to autostart agents
            :param agent_names: The names of the agents
            :param agent_map: The agent mapping to use
            :return: A string that contains the config file content.
        """
        environment_id = str(env.id)
        port: int = Config.get("server_rest_transport", "port", "8888")

        privatestatedir: str = os.path.join(Config.get("config", "state-dir", "/var/lib/inmanta"), environment_id)

        agent_deploy_splay: int
        agent_deploy_splay = yield env.get(data.AUTOSTART_AGENT_DEPLOY_SPLAY_TIME)
        agent_deploy_interval: int
        agent_deploy_interval = yield env.get(data.AUTOSTART_AGENT_DEPLOY_INTERVAL)

        agent_repair_splay: int
        agent_repair_splay = yield env.get(data.AUTOSTART_AGENT_REPAIR_SPLAY_TIME)
        agent_repair_interval: int
        agent_repair_interval = yield env.get(data.AUTOSTART_AGENT_REPAIR_INTERVAL)

        # generate config file
        config = """[config]
heartbeat-interval = 60
state-dir=%(statedir)s

agent-names = %(agents)s
environment=%(env_id)s
agent-map=%(agent_map)s

agent-deploy-splay-time=%(agent_deploy_splay)d
agent-deploy-interval=%(agent_deploy_interval)d
agent-repair-splay-time=%(agent_repair_splay)d
agent-repair-interval=%(agent_repair_interval)d

[agent_rest_transport]
port=%(port)s
host=%(serveradress)s
""" % {"agents": ",".join(agent_names), "env_id": environment_id, "port": port,
            "agent_map": ",".join(["%s=%s" % (k, v) for (k, v) in agent_map.items()]),
            "statedir": privatestatedir, "agent_deploy_splay": agent_deploy_splay,
            "agent_deploy_interval": agent_deploy_interval, "agent_repair_splay": agent_repair_splay,
            "agent_repair_interval": agent_repair_interval, "serveradress": server_config.server_address.get()}

        if server_config.server_enable_auth.get():
            token = encode_token(["agent"], environment_id)
            config += """
token=%s
    """ % (token)

        ssl_cert: Optional[str] = server_config.server_ssl_key.get()
        ssl_ca: Optional[str] = server_config.server_ssl_ca_cert.get()

        if ssl_ca is not None and ssl_cert is not None:
            # override CA
            config += """
ssl=True
ssl_ca_cert_file=%s
    """ % (ssl_ca)
        elif ssl_cert is not None:
            # system CA
            config += """
ssl=True
    """

        return config

    # Parameters

    @gen.coroutine
    def _request_parameter(self, env_id: uuid.UUID, resource_id: str) -> Apireturn:
        """
            Request the value of a parameter from an agent
        """
        if resource_id is not None and resource_id != "":
            env = yield data.Environment.get_by_id(env_id)

            # get a resource version
            res = yield data.Resource.get_latest_version(env_id, resource_id)

            if res is None:
                return 404, {"message": "The resource has no recent version."}

            rid: Id = Id.parse_id(res.resource_version_id)
            version: int = rid.version

            # only request facts of a resource every _fact_resource_block time
            now = time.time()
            if (resource_id not in self._fact_resource_block_set
                    or (self._fact_resource_block_set[resource_id] + self._fact_resource_block) < now):

                agents = yield data.ConfigurationModel.get_agents(env.id, version)
                yield self._ensure_agents(env, agents)

                client = self.get_agent_client(env_id, res.agent)
                if client is not None:
                    future = client.get_parameter(str(env_id), res.agent, res.to_dict())
                    self.add_future(future)

                self._fact_resource_block_set[resource_id] = now

            else:
                LOGGER.debug("Ignore fact request for %s, last request was sent %d seconds ago.",
                             resource_id, now - self._fact_resource_block_set[resource_id])

            return 503, {"message": "Agents queried for resource parameter."}
        else:
            return 404, {"message": "resource_id parameter is required."}

    @gen.coroutine
    def start_agents(self) -> NoneGen:
        """
            Ensure that autostarted agents of each environment are started when AUTOSTART_ON_START is true. This method
            is called on server start.
        """
        environments = yield data.Environment.get_list()
        for env in environments:
            agents = yield data.Agent.get_list(environment=env.id)
            autostart = yield env.get(data.AUTOSTART_ON_START)
            if autostart:
                agent_list = [a.name for a in agents]
                yield self._ensure_agents(env, agent_list)

    @gen.coroutine
    def restart_agents(self, env: data.Environment) -> NoneGen:
        agents = yield data.Agent.get_list(environment=env.id)
        autostart = yield env.get(data.AUTOSTART_ON_START)
        if autostart:
            agent_list = [a.name for a in agents]
            yield self._ensure_agents(env, agent_list, True)

    @gen.coroutine
    def stop_agents(self, env: data.Environment) -> NoneGen:
        """
            Stop all agents for this environment and close sessions
        """
        LOGGER.debug("Stopping all autostarted agents for env %s", env.id)
        if env.id in self._agent_procs:
            subproc = self._agent_procs[env.id]
            subproc.proc.terminate()
            yield wait_for_proc_bounded([subproc])
            del self._agent_procs[env.id]

        LOGGER.debug("Expiring all sessions for %s", env.id)
        sessions: List[protocol.Session]
        sessions = yield self.get_environment_sessions(env.id)
        for session in sessions:
            session.expire(0)



"""
    Copyright 2016 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import logging
import sys
import uuid

import colorlog
from inmanta import data
import pytest
from tornado.gen import sleep
from utils import retry_limited
from inmanta.server.protocol import Server, SessionListener, ServerSlice
from inmanta.server import SLICE_SESSION_MANAGER
from inmanta.protocol.methods import ENV_OPTS
from inmanta.protocol import method

LOGGER = logging.getLogger(__name__)


@method(method_name="status", operation="GET", index=True)
def get_status_x(tid: uuid.UUID):
    pass


@method(method_name="status", operation="GET", id=True, server_agent=True, timeout=10)
def get_agent_status_x(id):
    pass


# Methods need to be defined before the Client class is loaded by Python
from inmanta import protocol  # NOQA


class SessionSpy(SessionListener, ServerSlice):

    def __init__(self):
        ServerSlice.__init__(self, "sessionspy")
        self.expires = 0
        self.__sessions = []

    def new_session(self, session):
        self.__sessions.append(session)

    @protocol.handle(get_status_x)
    async def get_status_x(self, tid):
        status_list = []
        for session in self.__sessions:
            client = session.get_client()
            status = await client.get_agent_status_x("x")
            if status is not None and status.code == 200:
                status_list.append(status.result)

        return 200, {"agents": status_list}

    def expire(self, session, timeout):
        self.__sessions.remove(session)
        print(session._sid)
        self.expires += 1

    def get_sessions(self):
        return self.__sessions


class Agent(protocol.SessionEndpoint):

    @protocol.handle(get_agent_status_x)
    async def get_agent_status_x(self, id):
        return 200, {"status": "ok", "agents": self.end_point_names}


async def get_environment(env: uuid.UUID, metadata: dict):
    return data.Environment(from_mongo=True, _id=env, name="test", project=env, repo_url="xx", repo_branch="xx")


@pytest.mark.asyncio
async def test_2way_protocol(unused_tcp_port, logs=False):

    from inmanta.config import Config

    import inmanta.agent.config  # noqa: F401
    import inmanta.server.config  # noqa: F401

    if logs:
        # set logging to sensible defaults
        formatter = colorlog.ColoredFormatter(
            "%(log_color)s%(levelname)-8s%(reset)s %(green)s%(name)s %(blue)s%(message)s",
            datefmt=None,
            reset=True,
            log_colors={"DEBUG": "cyan", "INFO": "green", "WARNING": "yellow", "ERROR": "red", "CRITICAL": "red"},
        )

        stream = logging.StreamHandler()
        stream.setLevel(logging.DEBUG)

        if hasattr(sys.stdout, "isatty") and sys.stdout.isatty():
            stream.setFormatter(formatter)

        logging.root.handlers = []
        logging.root.addHandler(stream)
        logging.root.setLevel(logging.DEBUG)

    free_port = str(unused_tcp_port)
    Config.load_config()
    Config.set("server_rest_transport", "port", free_port)
    Config.set("agent_rest_transport", "port", free_port)
    Config.set("compiler_rest_transport", "port", free_port)
    Config.set("client_rest_transport", "port", free_port)
    Config.set("cmdline_rest_transport", "port", free_port)

    # Disable validation of envs
    old_get_env = ENV_OPTS["tid"].getter
    ENV_OPTS["tid"].getter = get_environment

    try:
        rs = Server()
        server = SessionSpy()
        rs.get_slice(SLICE_SESSION_MANAGER).add_listener(server)
        rs.add_slice(server)
        await rs.start()

        agent = Agent("agent")
        agent.add_end_point_name("agent")
        agent.set_environment(uuid.uuid4())
        await agent.start()

        await retry_limited(lambda: len(server.get_sessions()) == 1, 0.1)
        assert len(server.get_sessions()) == 1

        client = protocol.Client("client")
        status = await client.get_status_x(str(agent.environment))
        assert status.code == 200
        assert "agents" in status.result
        assert len(status.result["agents"]) == 1
        assert status.result["agents"][0]["status"], "ok"
        await server.stop()

        await rs.stop()
        await agent.stop()
    finally:
        ENV_OPTS["tid"].getter = old_get_env


async def check_sessions(sessions):
    for s in sessions:
        a = await s.client.get_agent_status_x("X")
        assert a.get_result()["status"] == "ok"


@pytest.mark.slowtest
@pytest.mark.asyncio(timeout=30)
async def test_timeout(unused_tcp_port):

    from inmanta.config import Config
    import inmanta.agent.config  # noqa: F401
    import inmanta.server.config  # noqa: F401

    free_port = str(unused_tcp_port)

    free_port = str(unused_tcp_port)

    # start server
    Config.load_config()
    Config.set("server_rest_transport", "port", free_port)
    Config.set("agent_rest_transport", "port", free_port)
    Config.set("compiler_rest_transport", "port", free_port)
    Config.set("client_rest_transport", "port", free_port)
    Config.set("cmdline_rest_transport", "port", free_port)
    Config.set("server", "agent-timeout", "1")

    # Disable validation of envs
    old_get_env = ENV_OPTS["tid"].getter
    ENV_OPTS["tid"].getter = get_environment

    try:

        rs = Server()
        server = SessionSpy()
        rs.get_slice(SLICE_SESSION_MANAGER).add_listener(server)
        rs.add_slice(server)
        await rs.start()

        env = uuid.uuid4()

        # agent 1
        agent = Agent("agent")
        agent.add_end_point_name("agent")
        agent.set_environment(env)
        await agent.start()

        # wait till up
        await retry_limited(lambda: len(server.get_sessions()) == 1, 0.1)
        assert len(server.get_sessions()) == 1

        # agent 2
        agent2 = Agent("agent")
        agent2.add_end_point_name("agent")
        agent2.set_environment(env)
        await agent2.start()

        # wait till up
        await retry_limited(lambda: len(server.get_sessions()) == 2, 0.1)
        assert len(server.get_sessions()) == 2

        # see if it stays up
        await check_sessions(server.get_sessions())
        await sleep(2)
        assert len(server.get_sessions()) == 2
        await check_sessions(server.get_sessions())

        # take it down
        await agent2.stop()

        # timout
        await sleep(2)
        # check if down
        assert len(server.get_sessions()) == 1
        print(server.get_sessions())
        await check_sessions(server.get_sessions())
        assert server.expires == 1
        await agent.stop()
        await server.stop()

        await rs.stop()
        await agent.stop()
    finally:
        ENV_OPTS["tid"].getter = old_get_env



"""
    Copyright 2016 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import os
import tempfile
import random
import string
import shutil
from tempfile import mktemp
import socket
import logging


import pytest

import utils
from inmanta import config, data, mongoproc
import inmanta.compiler as compiler
import pymongo
from motor import motor_asyncio
from inmanta.module import Project
from inmanta import resources, export
import inmanta.agent
from inmanta.agent import handler
from inmanta.agent.agent import Agent
from inmanta.ast import CompilerException
from click import testing
import inmanta.main
import re
from inmanta.server.bootloader import InmantaBootloader
from inmanta.server import SLICE_AGENT_MANAGER
from inmanta.export import cfg_env, unknown_parameters
import traceback
from tornado import process
import asyncio
from tornado.platform.asyncio import AnyThreadEventLoopPolicy
import sys
import pkg_resources
from typing import Optional, Dict
from inmanta import protocol


asyncio.set_event_loop_policy(AnyThreadEventLoopPolicy())

DEFAULT_PORT_ENVVAR = 'MONGOBOX_PORT'


@pytest.fixture(scope="session", autouse=True)
def mongo_db():
    db_path = tempfile.mkdtemp(dir="/dev/shm")
    mproc = mongoproc.MongoProc(db_path=db_path, port=get_free_tcp_port())
    port_envvar = DEFAULT_PORT_ENVVAR

    mproc.start()
    os.environ[port_envvar] = str(mproc.port)

    yield mproc

    mproc.stop()
    del os.environ[port_envvar]
    shutil.rmtree(db_path)


@pytest.fixture(scope="function", autouse=True)
def deactive_venv():
    old_os_path = os.environ.get("PATH", "")
    old_prefix = sys.prefix
    old_path = sys.path

    yield

    os.environ["PATH"] = old_os_path
    sys.prefix = old_prefix
    sys.path = old_path
    pkg_resources.working_set = pkg_resources.WorkingSet._build_master()


def reset_all():
    resources.resource.reset()
    export.Exporter.reset()
    process.Subprocess.uninitialize()
    asyncio.set_child_watcher(None)
    # No dynamic loading of commands at the moment, so no need to reset/reload
    # command.Commander.reset()
    handler.Commander.reset()
    Project._project = None
    unknown_parameters.clear()


@pytest.fixture(scope="function", autouse=True)
def clean_reset(mongo_client):
    cwd = os.getcwd()

    reset_all()
    yield
    reset_all()

    # reset cwd
    os.chdir(cwd)

    for db_name in mongo_client.list_database_names():
        if db_name != "admin":
            try:
                mongo_client.drop_database(db_name)
            except Exception:
                pass


@pytest.fixture(scope="session")
def mongo_client(mongo_db):
    '''Returns an instance of :class:`pymongo.MongoClient` connected
    to MongoBox database instance.
    '''
    port = int(mongo_db.port)
    return pymongo.MongoClient(port=port)


@pytest.fixture(scope="function")
def motor(mongo_db, mongo_client, event_loop):
    """
    Event_loop argument ensures this fixture is started after the eventloop is started
    """
    client = motor_asyncio.AsyncIOMotorClient('localhost', int(mongo_db.port))
    db = client["inmanta"]
    yield db


@pytest.fixture(scope="function")
async def data_module(motor):
    data.use_motor(motor)
    await data.create_indexes()


@pytest.fixture(scope="function")
def no_agent_backoff():
    backoff = inmanta.agent.agent.GET_RESOURCE_BACKOFF
    inmanta.agent.agent.GET_RESOURCE_BACKOFF = 0
    yield
    inmanta.agent.agent.GET_RESOURCE_BACKOFF = backoff


def get_free_tcp_port():
    """
        Semi safe method for getting a random port. This may contain a race condition.
    """
    tcp = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    tcp.bind(('', 0))
    _addr, port = tcp.getsockname()
    tcp.close()
    return str(port)


@pytest.fixture()
def free_port():
    port = get_free_tcp_port()
    yield port


@pytest.fixture(scope="function", autouse=True)
def inmanta_config():
    config.Config.load_config()
    config.Config.set("auth_jwt_default", "algorithm", "HS256")
    config.Config.set("auth_jwt_default", "sign", "true")
    config.Config.set("auth_jwt_default", "client_types", "agent,compiler")
    config.Config.set("auth_jwt_default", "key", "rID3kG4OwGpajIsxnGDhat4UFcMkyFZQc1y3oKQTPRs")
    config.Config.set("auth_jwt_default", "expire", "0")
    config.Config.set("auth_jwt_default", "issuer", "https://localhost:8888/")
    config.Config.set("auth_jwt_default", "audience", "https://localhost:8888/")

    yield config.Config._get_instance()
    config.Config._reset()


@pytest.fixture(scope="function")
async def agent_multi(server_multi, environment_multi):
    agentmanager = server_multi.get_slice(SLICE_AGENT_MANAGER)

    config.Config.set("config", "agent-deploy-interval", "0")
    config.Config.set("config", "agent-repair-interval", "0")
    a = Agent(hostname="node1", environment=environment_multi, agent_map={"agent1": "localhost"}, code_loader=False)
    a.add_end_point_name("agent1")
    await a.start()
    await utils.retry_limited(lambda: len(agentmanager.sessions) == 1, 10)

    yield a

    await a.stop()


@pytest.fixture(scope="function")
async def agent(server, environment):
    agentmanager = server.get_slice(SLICE_AGENT_MANAGER)

    config.Config.set("config", "agent-deploy-interval", "0")
    config.Config.set("config", "agent-repair-interval", "0")
    a = Agent(hostname="node1", environment=environment, agent_map={"agent1": "localhost"}, code_loader=False)
    a.add_end_point_name("agent1")
    await a.start()
    await utils.retry_limited(lambda: len(agentmanager.sessions) == 1, 10)

    yield a

    await a.stop()


@pytest.fixture(scope="function")
async def server(inmanta_config, mongo_db, mongo_client, motor):
    # fix for fact that pytest_tornado never set IOLoop._instance, the IOLoop of the main thread
    # causes handler failure

    state_dir = tempfile.mkdtemp()

    port = get_free_tcp_port()
    config.Config.set("database", "name", "inmanta-" + ''.join(random.choice(string.ascii_letters) for _ in range(10)))
    config.Config.set("database", "host", "localhost")
    config.Config.set("database", "port", str(mongo_db.port))
    config.Config.set("config", "state-dir", state_dir)
    config.Config.set("config", "log-dir", os.path.join(state_dir, "logs"))
    config.Config.set("server_rest_transport", "port", port)
    config.Config.set("agent_rest_transport", "port", port)
    config.Config.set("compiler_rest_transport", "port", port)
    config.Config.set("client_rest_transport", "port", port)
    config.Config.set("cmdline_rest_transport", "port", port)
    config.Config.set("config", "executable", os.path.abspath(os.path.join(__file__, "../../src/inmanta/app.py")))
    config.Config.set("server", "agent-timeout", "10")

    data.use_motor(motor)
    await data.create_indexes()

    ibl = InmantaBootloader()
    await ibl.start()

    yield ibl.restserver

    await ibl.stop()
    shutil.rmtree(state_dir)


@pytest.fixture(scope="function",
                params=[(True, True, False), (True, False, False), (False, True, False),
                        (False, False, False), (True, True, True)],
                ids=["SSL and Auth", "SSL", "Auth", "Normal", "SSL and Auth with not self signed certificate"])
async def server_multi(inmanta_config, mongo_db, mongo_client, request, motor):
    state_dir = tempfile.mkdtemp()

    ssl, auth, ca = request.param

    path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")

    if auth:
        config.Config.set("server", "auth", "true")

    for x, ct in [("server", None),
                  ("server_rest_transport", None),
                  ("agent_rest_transport", ["agent"]),
                  ("compiler_rest_transport", ["compiler"]),
                  ("client_rest_transport", ["api", "compiler"]),
                  ("cmdline_rest_transport", ["api"])]:
        if ssl and not ca:
            config.Config.set(x, "ssl_cert_file", os.path.join(path, "server.crt"))
            config.Config.set(x, "ssl_key_file", os.path.join(path, "server.open.key"))
            config.Config.set(x, "ssl_ca_cert_file", os.path.join(path, "server.crt"))
            config.Config.set(x, "ssl", "True")
        if ssl and ca:
            capath = os.path.join(path, "ca", "enduser-certs")

            config.Config.set(x, "ssl_cert_file", os.path.join(capath, "server.crt"))
            config.Config.set(x, "ssl_key_file", os.path.join(capath, "server.key.open"))
            config.Config.set(x, "ssl_ca_cert_file", os.path.join(capath, "server.chain"))
            config.Config.set(x, "ssl", "True")
        if auth and ct is not None:
            token = protocol.encode_token(ct)
            config.Config.set(x, "token", token)

    port = get_free_tcp_port()
    config.Config.set("database", "name", "inmanta-" + ''.join(random.choice(string.ascii_letters) for _ in range(10)))
    config.Config.set("database", "host", "localhost")
    config.Config.set("database", "port", str(mongo_db.port))
    config.Config.set("config", "state-dir", state_dir)
    config.Config.set("config", "log-dir", os.path.join(state_dir, "logs"))
    config.Config.set("server_rest_transport", "port", port)
    config.Config.set("agent_rest_transport", "port", port)
    config.Config.set("compiler_rest_transport", "port", port)
    config.Config.set("client_rest_transport", "port", port)
    config.Config.set("cmdline_rest_transport", "port", port)
    config.Config.set("config", "executable", os.path.abspath(os.path.join(__file__, "../../src/inmanta/app.py")))
    config.Config.set("server", "agent-timeout", "2")

    data.use_motor(motor)
    await data.create_indexes()

    ibl = InmantaBootloader()
    await ibl.start()

    yield ibl.restserver

    await ibl.stop()

    shutil.rmtree(state_dir)


@pytest.fixture(scope="function")
def client(server):
    client = protocol.Client("client")
    yield client


@pytest.fixture(scope="function")
def client_multi(server_multi):
    client = protocol.Client("client")
    yield client


@pytest.fixture(scope="function")
def sync_client_multi(server_multi):
    client = protocol.SyncClient("client")
    yield client


@pytest.fixture(scope="function", autouse=True)
def capture_warnings():
    logging.captureWarnings(True)
    yield
    logging.captureWarnings(False)


@pytest.fixture(scope="function")
async def environment(client, server):
    """
        Create a project and environment. This fixture returns the uuid of the environment
    """
    def create_project():
        return client.create_project("env-test")

    result = await create_project()
    assert(result.code == 200)
    project_id = result.result["project"]["id"]

    def create_env():
        return client.create_environment(project_id=project_id, name="dev")

    result = await create_env()
    env_id = result.result["environment"]["id"]

    cfg_env.set(env_id)

    yield env_id


@pytest.fixture(scope="function")
async def environment_multi(client_multi, server_multi):
    """
        Create a project and environment. This fixture returns the uuid of the environment
    """
    def create_project():
        return client_multi.create_project("env-test")

    result = await create_project()
    assert(result.code == 200)
    project_id = result.result["project"]["id"]

    def create_env():
        return client_multi.create_environment(project_id=project_id, name="dev")

    result = await create_env()
    env_id = result.result["environment"]["id"]

    yield env_id


class KeepOnFail(object):

    def keep(self) -> "Optional[Dict[str, str]]":
        pass


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    # execute all other hooks to obtain the report object
    outcome = yield
    rep = outcome.get_result()

    # we only look at actual failing test calls, not setup/teardown
    resources = {}
    if rep.when == "call" and rep.failed:
        for fixture in item.funcargs.values():
            if isinstance(fixture, KeepOnFail):
                msg = fixture.keep()
                if msg:
                    for label, res in msg.items():
                        resources[label] = res

        if resources:
            # we are behind report formatting, so write to report, not item
            rep.sections.append(("Resources Kept", "\n".join(
                ["%s %s" % (label, resource) for label, resource in resources.items()])))


async def off_main_thread(func):
    return await asyncio.get_event_loop().run_in_executor(None, func)


class SnippetCompilationTest(KeepOnFail):

    def setUpClass(self):
        self.libs = tempfile.mkdtemp()
        self.env = tempfile.mkdtemp()
        config.Config.load_config()
        self.cwd = os.getcwd()
        self.keep_shared = False

    def tearDownClass(self):
        if not self.keep_shared:
            shutil.rmtree(self.libs)
            shutil.rmtree(self.env)
        # reset cwd
        os.chdir(self.cwd)

    def setup_func(self):
        # init project
        self._keep = False
        self.project_dir = tempfile.mkdtemp()
        os.symlink(self.env, os.path.join(self.project_dir, ".env"))

    def tear_down_func(self):
        if not self._keep:
            shutil.rmtree(self.project_dir)

    def keep(self):
        self._keep = True
        self.keep_shared = True
        return {"env": self.env, "libs": self.libs, "project": self.project_dir}

    def setup_for_snippet(self, snippet, autostd=True):
        with open(os.path.join(self.project_dir, "project.yml"), "w") as cfg:
            cfg.write(
                """
            name: snippet test
            modulepath: [%s, %s]
            downloadpath: %s
            version: 1.0
            repo: ['https://github.com/inmanta/']"""
                % (self.libs,
                    os.path.join(os.path.dirname(os.path.abspath(__file__)), "data", "modules"),
                    self.libs))

        self.main = os.path.join(self.project_dir, "main.cf")
        with open(self.main, "w") as x:
            x.write(snippet)

        Project.set(Project(self.project_dir, autostd=autostd))

    def do_export(self, include_status=False, do_raise=True):
        return self._do_export(deploy=False, include_status=include_status, do_raise=do_raise)

    def _do_export(self, deploy=False, include_status=False, do_raise=True):
        """
        helper function to allow actual export to be run an a different thread
        i.e. export.run must run off main thread to allow it to start a new ioloop for run_sync
        """
        templfile = mktemp("json", "dump", self.project_dir)

        class Options(object):
            pass

        options = Options()
        options.json = templfile if not deploy else None
        options.depgraph = False
        options.deploy = deploy
        options.ssl = False

        from inmanta.export import Exporter  # noqa: H307

        try:
            (types, scopes) = compiler.do_compile()
        except Exception:
            types, scopes = (None, None)
            if do_raise:
                raise
            else:
                traceback.print_exc()

        # Even if the compile failed we might have collected additional data such as unknowns. So
        # continue the export
        export = Exporter(options)

        return export.run(types, scopes, model_export=False, include_status=include_status)

    async def do_export_and_deploy(self, include_status=False, do_raise=True):
        return await off_main_thread(lambda: self._do_export(deploy=True, include_status=include_status, do_raise=do_raise))

    def setup_for_error(self, snippet, shouldbe):
        self.setup_for_snippet(snippet)
        try:
            compiler.do_compile()
            assert False, "Should get exception"
        except CompilerException as e:
            text = e.format_trace(indent="  ")
            print(text)
            shouldbe = shouldbe.format(dir=self.project_dir)
            assert shouldbe == text

    def setup_for_error_re(self, snippet, shouldbe):
        self.setup_for_snippet(snippet)
        try:
            compiler.do_compile()
            assert False, "Should get exception"
        except CompilerException as e:
            text = e.format_trace(indent="  ")
            print(text)
            shouldbe = shouldbe.format(dir=self.project_dir)
            assert re.search(shouldbe, text) is not None


@pytest.fixture(scope="session")
def snippetcompiler_global():
    ast = SnippetCompilationTest()
    ast.setUpClass()
    yield ast
    ast.tearDownClass()


@pytest.fixture(scope="function")
def snippetcompiler(snippetcompiler_global):
    snippetcompiler_global.setup_func()
    yield snippetcompiler_global
    snippetcompiler_global.tear_down_func()


class CLI(object):
    async def run(self, *args):
        # set column width very wide so lines are not wrapped
        os.environ["COLUMNS"] = "1000"
        runner = testing.CliRunner()
        cmd_args = ["--host", "localhost", "--port", config.Config.get("cmdline_rest_transport", "port")]
        cmd_args.extend(args)

        def invoke():
            return runner.invoke(
                cli=inmanta.main.cmd,
                args=cmd_args,
                catch_exceptions=False
            )

        result = await asyncio.get_event_loop().run_in_executor(None, invoke)
        # reset to default again
        del os.environ["COLUMNS"]
        return result


@pytest.fixture
def cli():
    o = CLI()
    yield o



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import os
import glob
import imp
import hashlib
import json
import logging

import pkg_resources

VERSION_FILE = "version"
MODULE_DIR = "modules"
PERSIST_FILE = "modules.json"

LOGGER = logging.getLogger(__name__)


class CodeLoader(object):
    """
        Class responsible for managing code loaded from modules received from the compiler

        :param code_dir The directory where the code is stored
    """

    def __init__(self, code_dir):
        self.__code_dir = code_dir
        self.__modules = {}
        self.__current_version = 0

        self.__check_dir()
        self.load_modules()

    def load_modules(self):
        """
            Load all existing modules
        """
        mod_dir = os.path.join(self.__code_dir, MODULE_DIR)

        if os.path.exists(os.path.join(self.__code_dir, VERSION_FILE)):
            fd = open(os.path.join(self.__code_dir, VERSION_FILE), "r")
            self.__current_version = int(fd.read())
            fd.close()

        pkg_resources.working_set = pkg_resources.WorkingSet._build_master()

        for py in glob.glob(os.path.join(mod_dir, "*.py")):
            if mod_dir in py:
                mod_name = py[len(mod_dir) + 1:-3]
            else:
                mod_name = py[:-3]

            source_code = ""
            with open(py, "r") as fd:
                source_code = fd.read().encode("utf-8")

            sha1sum = hashlib.new("sha1")
            sha1sum.update(source_code)

            hv = sha1sum.hexdigest()

            self._load_module(mod_name, py, hv)

    def __check_dir(self):
        """
            Check if the code directory
        """
        # check for the code dir
        if not os.path.exists(self.__code_dir):
            os.makedirs(self.__code_dir, exist_ok=True)

        # check for modules subdir
        if not os.path.exists(os.path.join(self.__code_dir, MODULE_DIR)):
            os.makedirs(os.path.join(self.__code_dir, MODULE_DIR), exist_ok=True)

    def _load_module(self, mod_name, python_file, hv):
        """
            Load or reload a module
        """
        try:
            mod = imp.load_source(mod_name, python_file)
            self.__modules[mod_name] = (hv, mod)
            LOGGER.info("Loaded module %s" % mod_name)
        except ImportError:
            LOGGER.exception("Unable to load module %s" % mod_name)

    def deploy_version(self, key, mod, persist=False):
        """
            Deploy a new version of the modules

            :param version The version of the deployed modules
            :modules modules A list of module names and the hashes of the code files
        """
        # deploy the new code
        name = mod[1]
        source_code = mod[2]

        # if the module is new, or update
        if name not in self.__modules or key != self.__modules[name][0]:
            LOGGER.info("Deploying code (key=%s, module=%s)", key, mod[1])
            # write the new source
            source_file = os.path.join(self.__code_dir, MODULE_DIR, name + ".py")

            fd = open(source_file, "w+")
            fd.write(source_code)
            fd.close()

            # (re)load the new source
            self._load_module(name, source_file, key)

        if persist:
            with open(os.path.join(self.__code_dir, PERSIST_FILE), "w+") as fd:
                json.dump(mod, fd)

        pkg_resources.working_set = pkg_resources.WorkingSet._build_master()

    def get_module_payload(self):
        """
            Get the lastest module code payload in json formatted string
        """
        with open(os.path.join(self.__code_dir, PERSIST_FILE), "w+") as fd:
            return fd.read()
